"""Menu item that returns the user to the previous menu."""
import subprocess
from collections import defaultdict
from copy import deepcopy

from getch import pause
from halo import Halo

from vigorish.cli.components import (
    data_sets_prompt,
    file_types_prompt,
    get_random_cli_color,
    get_random_dots_spinner,
    print_heading,
    print_message,
    season_prompt,
    user_options_prompt,
)
from vigorish.cli.components.viewers import DictListTableViewer
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.enums import DataSet, SyncDirection, VigFile
from vigorish.tasks.sync_scraped_data import SyncScrapedDataTask
from vigorish.util.dt_format_strings import DT_NAIVE_LESS
from vigorish.util.result import Result
from vigorish.util.sys_helpers import file_size_str

SYNC_STATUS_TEXT_COLOR = {"out_of_sync": "bright_green", "in_sync": "blue"}


class SyncScrapedData(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.s3_sync = SyncScrapedDataTask(self.app)
        self.year = None
        self.sync_direction = ""
        self.file_types = []
        self.sync_tasks = defaultdict(list)
        self.sync_files = defaultdict(dict)
        self.sync_results = []
        self.task_number = 0
        self.spinners = defaultdict(dict)
        self.menu_item_text = " Synchronize Scraped Data"
        self.menu_item_emoji = EMOJIS.get("CLOUD", "")
        self.menu_heading = self.menu_item_text
        self.exit_menu = False

    @property
    def total_tasks(self):
        return sum(len(data_sets) for data_sets in self.sync_tasks.values())

    @property
    def all_files_are_in_sync(self):
        return all(
            not out_of_sync
            for doc_format_dict in self.sync_files.values()
            for (out_of_sync, _, _) in doc_format_dict.values()
        )

    def launch(self):
        result = self.get_sync_parameters()
        if result.failure:
            return Result.Ok(False)
        self.get_all_s3_objects()
        for file_type, data_sets in self.sync_tasks.items():
            for data_set in data_sets:
                result = self.find_out_of_sync_files(file_type, data_set)
                if result.failure:
                    return result
        self.report_sync_results()
        return self.synchronize_files()

    def get_sync_parameters(self):
        result = season_prompt(self.db_session, "Select a season to synchronize scraped data:")
        if result.failure:
            return Result.Fail("")
        self.year = result.value.year
        self.file_types = file_types_prompt("Select one or multiple file types to synchronize:")
        for file_type in self.file_types:
            self.sync_tasks[file_type] = self.get_data_sets_to_sync(file_type)
        result = self.sync_direction_prompt()
        if result.failure:
            return Result.Fail("")
        self.sync_direction = result.value
        return Result.Ok()

    def get_data_sets_to_sync(self, file_type):
        heading = self.get_menu_heading(f"Select Data Sets to Sync for {file_type}")
        prompt = f"Select all data sets to synchronize for {file_type} files"
        if file_type == VigFile.COMBINED_GAME_DATA:
            return [DataSet.ALL]
        valid_data_sets = (
            [DataSet.ALL]
            if file_type != VigFile.PATCH_LIST
            else [
                DataSet.BBREF_GAMES_FOR_DATE,
                DataSet.BBREF_BOXSCORES,
                DataSet.BROOKS_GAMES_FOR_DATE,
                DataSet.BROOKS_PITCHFX,
            ]
        )
        return data_sets_prompt(heading, prompt, valid_data_sets)

    def sync_direction_prompt(self):
        prompt = (
            'Please choose the "direction" of the sync operation. Files in the source location '
            "are compared to files in the destination, if a source file does not exist in the "
            "destination, it will be copied to the destination. Files that exist in both "
            "locations are compared, if the destination file is older it will be replaced by "
            "the updated version from the source location. Files are never deleted from either "
            "location during a sync."
        )
        SYNC_UP_MENU_CHOICE = "Sync Up: Local Folder (Source) -> S3 Bucket (Dest)"
        SYNC_DOWN_MENU_CHOICE = "Sync Down: S3 Bucket (Source) -> Local Folder (Dest)"
        choices = {
            f"{MENU_NUMBERS.get(1)}  {SYNC_UP_MENU_CHOICE}": SyncDirection.UP_TO_S3,
            f"{MENU_NUMBERS.get(2)}  {SYNC_DOWN_MENU_CHOICE}": SyncDirection.DOWN_TO_LOCAL,
            f"{EMOJIS.get('BACK')} Return to Main Menu": None,
        }
        return user_options_prompt(choices, prompt)

    def get_all_s3_objects(self):
        subprocess.run(["clear"])
        spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        spinner.text = "Retrieving details of all objects stored in S3..."
        spinner.start()
        self.s3_sync.get_all_s3_objects()
        spinner.stop()

    def find_out_of_sync_files(self, file_type, data_set):
        self.task_number += 1
        self.report_sync_results()
        self.update_spinner(file_type, data_set)
        result = Result.Fail("Failed to sync files")
        result = self.s3_sync.find_out_of_sync_files(self.sync_direction, file_type, data_set, self.year)
        if result.failure:
            self.spinners[file_type][data_set].stop()
            return result
        (out_of_sync, missing_files, update_files) = result.value
        self.sync_files[file_type][data_set] = (out_of_sync, missing_files, update_files)
        if out_of_sync:
            self.sync_results.append(self.out_of_sync_text(file_type, data_set, len(missing_files) + len(update_files)))
        else:
            self.sync_results.append(self.in_sync_text(file_type, data_set))
        self.spinners[file_type][data_set].stop()
        return Result.Ok()

    def report_sync_results(self):
        subprocess.run(["clear"])
        src_folder = "S3 bucket" if self.sync_direction == SyncDirection.DOWN_TO_LOCAL else "local folder"
        dest_folder = "local folder" if self.sync_direction == SyncDirection.DOWN_TO_LOCAL else "S3 bucket"
        print_heading(f"Syncing data from {src_folder} to {dest_folder}", fg="bright_yellow")
        for (result, text_color) in self.sync_results:
            print_message(result, fg=text_color)

    def update_spinner(self, file_type, data_set):
        self.spinners[file_type][data_set] = Halo(
            spinner=get_random_dots_spinner(),
            color=get_random_cli_color(),
            text=(
                f"Analyzing MLB {self.year} {data_set} {file_type} files "
                f"(Task {self.task_number}/{self.total_tasks})..."
            ),
        )
        self.spinners[file_type][data_set].start()

    def out_of_sync_text(self, file_type, data_set, sync_count):
        out_of_sync = (
            f"{sync_count} files out of sync: {self.year} {data_set} {file_type} "
            f"(Task {self.task_number}/{self.total_tasks})"
        )
        return (out_of_sync, SYNC_STATUS_TEXT_COLOR["out_of_sync"])

    def in_sync_text(self, file_type, data_set):
        in_sync = f"All files in sync: {self.year} {data_set} {file_type} (Task {self.task_number}/{self.total_tasks})"
        return (in_sync, SYNC_STATUS_TEXT_COLOR["in_sync"])

    def synchronize_files(self):
        if self.all_files_are_in_sync:
            message = "All files for selected data sets are in sync!"
            print_message(message, fg="bright_green", bold=True)
            pause(message="Press any key to continue...")
            return Result.Ok()
        for file_type, file_type_dict in self.sync_files.items():
            for data_set, (out_of_sync, missing_files, outdated_files) in file_type_dict.items():
                if not out_of_sync:
                    continue
                all_sync_files = []
                missing_count = 0
                outdated_count = 0
                if missing_files:
                    all_sync_files.extend(missing_files)
                    missing_count = len(missing_files)
                if outdated_files:
                    all_sync_files.extend(outdated_files)
                    outdated_count = len(outdated_files)
                table_viewer = self.create_table_viewer(
                    all_sync_files, data_set, file_type, missing_count, outdated_count
                )
                apply_changes = table_viewer.launch()
                if apply_changes:
                    self.apply_pending_changes(file_type, data_set, missing_files, outdated_files)
        return Result.Ok()

    def create_table_viewer(self, sync_files, data_set, file_type, missing_count, outdated_count):
        dict_list = [
            {
                "filename": f["name"],
                "size": file_size_str(f["size"]),
                "last_modified": f'{f["last_modified"].strftime(DT_NAIVE_LESS)} UTC',
                "sync_operation": f["operation"],
            }
            for f in sync_files
        ]
        m = []
        if missing_count:
            missing_plural = "files below do" if missing_count > 1 else "file below does"
            file_dest = "S3 bucket" if self.sync_direction == SyncDirection.UP_TO_S3 else "local folder"
            m.append(f"{missing_count} {missing_plural} not exist in the {file_dest}")
        if outdated_count:
            outdated_plural = "files have" if outdated_count > 1 else "file has"
            file_src = "local folder" if self.sync_direction == SyncDirection.UP_TO_S3 else "S3 bucket"
            m.append(f"{outdated_count} {outdated_plural} a more recent version in the {file_src}")
        return DictListTableViewer(
            dict_list,
            prompt="Would you like to apply the changes to the files listed above?",
            heading=f"File sync changes for {file_type} Files (Data Set: {data_set})",
            heading_color="bright_yellow",
            message="\n".join(m),
            table_color="bright_cyan",
        )

    def apply_pending_changes(self, file_type, data_set, missing_files, outdated_files):
        subprocess.run(["clear"])
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.spinner.start()
        self.s3_sync.events.sync_files_progress += self.update_sync_progress
        self.s3_sync.sync_files(self.sync_direction, missing_files, outdated_files, file_type, data_set, self.year)
        self.s3_sync.events.sync_files_progress -= self.update_sync_progress
        self.spinner.stop()
        src_folder = "S3 bucket" if self.sync_direction == SyncDirection.DOWN_TO_LOCAL else "local folder"
        dest_folder = "local folder" if self.sync_direction == SyncDirection.DOWN_TO_LOCAL else "S3 bucket"
        message = (
            f"All changes have been applied, MLB {self.year} {data_set} {file_type} files in {src_folder} "
            f"have been synced to {dest_folder}!"
        )
        print_message(message, fg="bright_green", bold=True)
        pause(message="Press any key to continue...")

    def get_sync_files(self, missing_files, outdated_files):
        sync_files = []
        if not missing_files and not outdated_files:
            return sync_files
        if missing_files:
            sync_files = deepcopy(missing_files)
        if outdated_files:
            sync_files.extend(deepcopy(outdated_files))
        return sync_files

    def update_sync_progress(self, name, complete, total):
        direction = "Down" if self.sync_direction == SyncDirection.DOWN_TO_LOCAL else "Up"
        percent = complete / float(total)
        self.spinner.text = f"{direction}loading {name} {percent:.0%} ({complete}/{total} Files)"



from functools import cached_property
from typing import Dict, List, Tuple, Union

from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session

import vigorish.database as db
from vigorish.cli.components.viewers import create_display_table, create_table_viewer
from vigorish.data.metrics import BatStatsMetrics, PitchFxMetricsCollection, PitchStatsMetrics
from vigorish.data.scraped_data import ScrapedData
from vigorish.enums import PitchType
from vigorish.util.exceptions import UnknownPlayerException
from vigorish.util.string_helpers import validate_pitch_app_id

PitchTypePercentiles = List[Dict[str, Union[PitchType, Tuple[float, float]]]]


class PlayerData:
    def __init__(self, app, mlb_id: int):
        self.app = app
        self.mlb_id = mlb_id
        self.db_engine: Engine = self.app.db_engine
        self.db_session: Session = self.app.db_session
        self.scraped_data: ScrapedData = self.app.scraped_data
        self.player = db.Player.find_by_mlb_id(self.db_session, self.mlb_id)
        if not self.player:
            raise UnknownPlayerException(mlb_id)
        self.player_id = db.PlayerId.find_by_mlb_id(self.db_session, self.mlb_id)
        self.pitchfx_game_data_cache = {}

    @property
    def player_name(self) -> str:
        return self.player_id.mlb_name

    @property
    def years_played(self) -> List[int]:
        return [s.year for s in self.seasons_played] if self.seasons_played else []

    @property
    def pitch_app_map(self) -> Dict[str, db.PitchAppScrapeStatus]:
        return {pitch_app.bbref_game_id: pitch_app for pitch_app in self.player.pitch_apps}

    @cached_property
    def seasons_played(self) -> List[db.Season]:
        return self.scraped_data.get_all_seasons_with_data_for_player(self.mlb_id)

    @cached_property
    def pitch_stats_for_career(self) -> PitchStatsMetrics:
        return self.scraped_data.get_pitch_stats_for_career_for_player(self.mlb_id)

    @cached_property
    def pitch_stats_as_sp(self) -> PitchStatsMetrics:
        return self.scraped_data.get_pitch_stats_as_sp_for_player(self.mlb_id)

    @cached_property
    def pitch_stats_as_rp(self) -> PitchStatsMetrics:
        return self.scraped_data.get_pitch_stats_as_rp_for_player(self.mlb_id)

    @cached_property
    def pitch_stats_by_year(self) -> List[PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_by_year_for_player(self.mlb_id)

    @cached_property
    def pitch_stats_by_team(self) -> List[PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_by_team_for_player(self.mlb_id)

    @cached_property
    def pitch_stats_by_team_by_year(self) -> List[PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_by_team_by_year_for_player(self.mlb_id)

    @cached_property
    def pitch_stats_by_opp_team(self) -> List[PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_by_opp_team_for_player(self.mlb_id)

    @cached_property
    def pitch_stats_by_opp_team_by_year(self) -> List[PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_by_opp_team_by_year_for_player(self.mlb_id)

    @cached_property
    def bat_stats_for_career(self) -> BatStatsMetrics:
        return self.scraped_data.get_bat_stats_for_career_for_player(self.mlb_id)

    @cached_property
    def bat_stats_by_year(self) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_by_year_for_player(self.mlb_id)

    @cached_property
    def bat_stats_by_team(self) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_by_team_for_player(self.mlb_id)

    @cached_property
    def bat_stats_by_team_by_year(self) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_by_team_by_year_for_player(self.mlb_id)

    @cached_property
    def bat_stats_by_opp_team(self) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_by_opp_for_player(self.mlb_id)

    @cached_property
    def bat_stats_by_opp_team_by_year(self) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_by_opp_by_year_for_player(self.mlb_id)

    @cached_property
    def pfx_pitching_metrics_for_career(self) -> PitchFxMetricsCollection:
        return self.scraped_data.get_career_pfx_metrics_for_pitcher(self.mlb_id)

    @cached_property
    def percentiles_for_pitch_types_for_career(self) -> List[PitchTypePercentiles]:
        return [
            self.scraped_data.calculate_pitch_type_percentiles(self.player.throws, pfx_metrics)
            for pfx_metrics in self.pfx_pitching_metrics_for_career.pitch_type_metrics.values()
        ]

    @cached_property
    def pfx_pitching_metrics_vs_rhb_for_career(self) -> PitchFxMetricsCollection:
        return self.scraped_data.get_career_pfx_metrics_vs_rhb_for_pitcher(self.mlb_id)

    @cached_property
    def percentiles_for_pitch_types_vs_rhb_for_career(self) -> List[PitchTypePercentiles]:
        return [
            self.scraped_data.calculate_pitch_type_percentiles(self.player.throws, pfx_metrics)
            for pfx_metrics in self.pfx_pitching_metrics_vs_rhb_for_career.pitch_type_metrics.values()
        ]

    @cached_property
    def pfx_pitching_metrics_vs_lhb_for_career(self) -> PitchFxMetricsCollection:
        return self.scraped_data.get_career_pfx_metrics_vs_lhb_for_pitcher(self.mlb_id)

    @cached_property
    def percentiles_for_pitch_types_vs_lhb_for_career(self) -> List[PitchTypePercentiles]:
        return [
            self.scraped_data.calculate_pitch_type_percentiles(self.player.throws, pfx_metrics)
            for pfx_metrics in self.pfx_pitching_metrics_vs_lhb_for_career.pitch_type_metrics.values()
        ]

    @cached_property
    def pfx_pitching_metrics_by_year(self) -> Dict[int, PitchFxMetricsCollection]:
        return {
            s.year: self.scraped_data.get_pfx_metrics_for_year_for_pitcher(self.mlb_id, s.year)
            for s in self.seasons_played
        }

    @cached_property
    def percentiles_for_pitch_types_by_year(self) -> Dict[int, List[PitchTypePercentiles]]:
        return {
            year: [
                self.scraped_data.calculate_pitch_type_percentiles(self.player.throws, pfx_metrics)
                for pfx_metrics in pfx_metrics_collection.pitch_type_metrics.values()
            ]
            for year, pfx_metrics_collection in self.pfx_pitching_metrics_by_year.items()
        }

    @cached_property
    def pfx_pitching_metrics_vs_rhb_by_year(self) -> Dict[int, PitchFxMetricsCollection]:
        return {
            s.year: self.scraped_data.get_pfx_metrics_for_year_vs_rhb_for_pitcher(self.mlb_id, s.year)
            for s in self.seasons_played
        }

    @cached_property
    def percentiles_for_pitch_types_vs_rhb_by_year(
        self,
    ) -> Dict[int, List[PitchTypePercentiles]]:
        return {
            year: [
                self.scraped_data.calculate_pitch_type_percentiles(self.player.throws, pfx_metrics)
                for pfx_metrics in pfx_metrics_collection.pitch_type_metrics.values()
            ]
            for year, pfx_metrics_collection in self.pfx_pitching_metrics_vs_rhb_by_year.items()
        }

    @cached_property
    def pfx_pitching_metrics_vs_lhb_by_year(self) -> Dict[int, PitchFxMetricsCollection]:
        return {
            s.year: self.scraped_data.get_pfx_metrics_for_year_vs_lhb_for_pitcher(self.mlb_id, s.year)
            for s in self.seasons_played
        }

    @cached_property
    def percentiles_for_pitch_types_vs_lhb_by_year(
        self,
    ) -> Dict[int, List[PitchTypePercentiles]]:
        return {
            year: [
                self.scraped_data.calculate_pitch_type_percentiles(self.player.throws, pfx_metrics)
                for pfx_metrics in pfx_metrics_collection.pitch_type_metrics.values()
            ]
            for year, pfx_metrics_collection in self.pfx_pitching_metrics_vs_lhb_by_year.items()
        }

    @cached_property
    def pfx_batting_metrics_for_career(self) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_for_career_for_batter(self.mlb_id)

    @cached_property
    def pfx_batting_metrics_vs_rhp_as_rhb_for_career(self) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_vs_rhp_as_rhb_for_career_for_batter(self.mlb_id)

    @cached_property
    def pfx_batting_metrics_vs_rhp_as_lhb_for_career(self) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_vs_rhp_as_lhb_for_career_for_batter(self.mlb_id)

    @cached_property
    def pfx_batting_metrics_vs_lhp_as_lhb_for_career(self) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_vs_lhp_as_lhb_for_career_for_batter(self.mlb_id)

    @cached_property
    def pfx_batting_metrics_vs_lhp_as_rhb_for_career(self) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_vs_lhp_as_rhb_for_career_for_batter(self.mlb_id)

    @cached_property
    def pfx_batting_metrics_by_year(self) -> Dict[int, PitchFxMetricsCollection]:
        return {
            s.year: self.scraped_data.get_pfx_metrics_for_year_for_batter(self.mlb_id, s.year)
            for s in self.seasons_played
        }

    def get_pfx_pitching_metrics_for_game(self, bbref_game_id: str) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_for_game_for_pitcher(self.mlb_id, bbref_game_id)

    def get_pfx_pitching_metrics_vs_rhb_for_game(self, bbref_game_id: str) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_for_game_vs_rhb_for_pitcher(self.mlb_id, bbref_game_id)

    def get_pfx_pitching_metrics_vs_lhb_for_game(self, bbref_game_id: str) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_for_game_vs_lhb_for_pitcher(self.mlb_id, bbref_game_id)

    def get_pfx_batting_metrics_for_game(self, bbref_game_id) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_for_game_for_batter(self.mlb_id, bbref_game_id)

    def get_pfx_batting_metrics_vs_rhp_as_rhb_for_game(self, bbref_game_id) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_vs_rhp_as_rhb_for_game_for_batter(self.mlb_id, bbref_game_id)

    def get_pfx_batting_metrics_vs_rhp_as_lhb_for_game(self, bbref_game_id) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_vs_rhp_as_lhb_for_game_for_batter(self.mlb_id, bbref_game_id)

    def get_pfx_batting_metrics_vs_lhp_as_lhb_for_game(self, bbref_game_id) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_vs_lhp_as_lhb_for_game_for_batter(self.mlb_id, bbref_game_id)

    def get_pfx_batting_metrics_vs_lhp_as_rhb_for_game(self, bbref_game_id) -> PitchFxMetricsCollection:
        return self.scraped_data.get_pfx_metrics_vs_lhp_as_rhb_for_game_for_batter(self.mlb_id, bbref_game_id)

    def get_pfx_metrics_collection_for_game(self, bbref_game_id=None, pitch_app_id=None):
        pitch_app = self.get_pitch_app(bbref_game_id, pitch_app_id)
        if not pitch_app:
            return None
        pfx_metrics_for_game = self.pitchfx_game_data_cache.get(bbref_game_id)
        if not pfx_metrics_for_game:
            pfx_metrics_for_game = {
                "all": self.get_pfx_pitching_metrics_for_game(bbref_game_id),
                "vs_rhb": self.get_pfx_pitching_metrics_vs_rhb_for_game(bbref_game_id),
                "vs_lhb": self.get_pfx_pitching_metrics_vs_lhb_for_game(bbref_game_id),
            }
            self.pitchfx_game_data_cache[bbref_game_id] = pfx_metrics_for_game
        return pfx_metrics_for_game

    def get_pitch_app(self, bbref_game_id, pitch_app_id):
        if not pitch_app_id and not bbref_game_id:
            return None
        if not bbref_game_id:
            bbref_game_id = validate_pitch_app_id(pitch_app_id).value["game_id"]
        return self.pitch_app_map.get(bbref_game_id)

    def view_pitch_mix_batter_stance_splits(self, game_id):
        return create_table_viewer(
            [
                create_display_table(
                    table_rows=self.get_pitch_mix_batter_stance_splits_for_game(game_id),
                    heading=f"Pitch Mix for {self.player_name} by Batter Stance (This Game)",
                    table_headers="keys",
                ),
                create_display_table(
                    table_rows=self.get_pitch_mix_batter_stance_splits_for_career(),
                    heading=f"Pitch Mix for {self.player_name} by Batter Stance (Career)",
                    table_headers="keys",
                ),
            ]
        )

    def get_pitch_mix_batter_stance_splits_for_game(self, game_id):
        pitch_metrics = self.get_pfx_metrics_collection_for_game(bbref_game_id=game_id)
        return self.get_pitch_mix_batter_stance_splits(
            pitch_metrics["all"], pitch_metrics["vs_rhb"], pitch_metrics["vs_lhb"], include_pitch_count=True
        )

    def get_pitch_mix_batter_stance_splits_for_career(self):
        return self.get_pitch_mix_batter_stance_splits(
            self.pfx_pitching_metrics_for_career,
            self.pfx_pitching_metrics_vs_rhb_for_career,
            self.pfx_pitching_metrics_vs_lhb_for_career,
        )

    def get_pitch_mix_batter_stance_splits(
        self, pitch_metrics_all, pitch_metrics_vs_rhb, pitch_metrics_vs_lhb, include_pitch_count=False
    ):
        return [
            {
                "pitch_type": pitch_type.print_name,
                "all": pitch_metrics_all.get_usage_stats_for_pitch_type(pitch_type, include_pitch_count),
                "vs_RHB": pitch_metrics_vs_rhb.get_usage_stats_for_pitch_type(pitch_type, include_pitch_count),
                "vs_LHB": pitch_metrics_vs_lhb.get_usage_stats_for_pitch_type(pitch_type, include_pitch_count),
            }
            for pitch_type in pitch_metrics_all.pitch_types
        ]

    def view_pitch_mix_season_splits(self, game_id):
        return create_table_viewer(
            [
                create_display_table(
                    table_rows=self.get_pitch_mix_season_splits(game_id),
                    heading=f"Pitch Mix for {self.player_name} by Season",
                    table_headers="keys",
                )
            ]
        )

    def get_pitch_mix_season_splits(self, game_id):
        return [
            self.get_pitch_mix_season_splits_for_pitch_type(game_id, pitch_type)
            for pitch_type in self.pfx_pitching_metrics_for_career.pitch_types
        ]

    def get_pitch_mix_season_splits_for_pitch_type(self, game_id, pitch_type):
        pitch_metrics_for_game = self.get_pfx_metrics_collection_for_game(bbref_game_id=game_id)
        table_row = {
            "pitch_type": pitch_type.print_name,
            "game": pitch_metrics_for_game["all"].get_usage_stats_for_pitch_type(pitch_type),
            "all": self.pfx_pitching_metrics_for_career.get_usage_stats_for_pitch_type(pitch_type),
        }
        for year, pitch_metrics_for_year in self.pfx_pitching_metrics_by_year.items():
            table_row[str(year)] = pitch_metrics_for_year.get_usage_stats_for_pitch_type(pitch_type)
        return table_row

    def view_plate_discipline_pitch_type_splits(self, game_id):
        return create_table_viewer(
            [
                create_display_table(
                    table_rows=self.get_plate_discipline_pitch_type_splits_for_game(game_id),
                    heading=f"Plate Discipline Stats for {self.player_name} (This Game)",
                    table_headers="keys",
                ),
                create_display_table(
                    table_rows=self.get_plate_discipline_pitch_type_splits_for_career(),
                    heading=f"Plate Discipline Stats for {self.player_name} (Career)",
                    table_headers="keys",
                ),
            ]
        )

    def get_plate_discipline_pitch_type_splits_for_game(self, game_id):
        pitch_metrics = self.get_pfx_metrics_collection_for_game(bbref_game_id=game_id)
        return pitch_metrics["all"].get_plate_discipline_pitch_type_splits(include_pitch_count=True)

    def get_plate_discipline_pitch_type_splits_for_career(self):
        return self.pfx_pitching_metrics_for_career.get_plate_discipline_pitch_type_splits()

    def view_batted_ball_pitch_type_splits(self, game_id):
        return create_table_viewer(
            [
                create_display_table(
                    table_rows=self.get_batted_ball_pitch_type_splits_for_game(game_id),
                    heading=f"Batted Ball Stats for {self.player_name} (This Game)",
                    table_headers="keys",
                ),
                create_display_table(
                    table_rows=self.get_batted_ball_pitch_type_splits_for_career(),
                    heading=f"Batted Ball Stats for {self.player_name} (Career)",
                    table_headers="keys",
                ),
            ]
        )

    def get_batted_ball_pitch_type_splits_for_game(self, game_id):
        pitch_metrics = self.get_pfx_metrics_collection_for_game(bbref_game_id=game_id)
        return pitch_metrics["all"].get_batted_ball_pitch_type_splits(include_bip_count=True)

    def get_batted_ball_pitch_type_splits_for_career(self):
        return self.pfx_pitching_metrics_for_career.get_batted_ball_pitch_type_splits()

    def view_bat_stats_pitch_type_splits(self, game_id):
        return create_table_viewer(
            [
                create_display_table(
                    table_rows=self.get_bat_stats_pitch_type_splits_for_game(game_id),
                    heading=f"Batting Stats for {self.player_name} (This Game)",
                    table_headers="keys",
                ),
                create_display_table(
                    table_rows=self.get_bat_stats_pitch_type_splits_for_career(),
                    heading=f"Batting Stats for {self.player_name} (Career)",
                    table_headers="keys",
                ),
            ]
        )

    def get_bat_stats_pitch_type_splits_for_game(self, game_id):
        pitch_metrics = self.get_pfx_metrics_collection_for_game(bbref_game_id=game_id)
        return pitch_metrics["all"].get_bat_stats_pitch_type_splits(include_pa_count=True)

    def get_bat_stats_pitch_type_splits_for_career(self):
        return self.pfx_pitching_metrics_for_career.get_bat_stats_pitch_type_splits()



# flake8: noqa
from vigorish.models.bat_stats import BatStats, BatStatsCsvRow
from vigorish.models.pitch_stats import PitchStats, PitchStatsCsvRow
from vigorish.models.pitch_type_percentiles import PitchTypePercentile
from vigorish.models.pitchfx import PitchFx, PitchFxCsvRow
from vigorish.models.player import Player
from vigorish.models.player_id import PlayerId
from vigorish.models.player_team_assoc import Assoc_Player_Team
from vigorish.models.scrape_error import ScrapeError
from vigorish.models.scrape_job import ScrapeJob
from vigorish.models.season import Season
from vigorish.models.status_date import DateScrapeStatus, DateScrapeStatusCsvRow
from vigorish.models.status_game import GameScrapeStatus, GameScrapeStatusCsvRow
from vigorish.models.status_pitch_appearance import PitchAppScrapeStatus, PitchAppScrapeStatusCsvRow
from vigorish.models.team import Team
from vigorish.models.time_between_pitches import TimeBetweenPitches



"""Record of a job to scrape data for a specified date range."""
from datetime import timezone
from uuid import uuid4

from sqlalchemy import Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy.types import Enum

import vigorish.database as db
from vigorish.config.project_paths import NODEJS_INBOX, NODEJS_OUTBOX
from vigorish.constants import DATA_SET_TO_NAME_MAP
from vigorish.enums import DataSet, JobStatus
from vigorish.util.datetime_util import (
    get_date_range,
    get_local_utcoffset,
    localized_dt_string,
    make_tzaware,
    utc_now,
)
from vigorish.util.dt_format_strings import DATE_ONLY
from vigorish.util.list_helpers import group_and_sort_list
from vigorish.util.string_helpers import string_is_null_or_blank


class ScrapeJob(db.Base):
    """Record of a job to scrape data for a specified date range."""

    __tablename__ = "scrape_job"
    id = Column(String, primary_key=True, default=lambda: f"{str(uuid4())[:4]}")
    name = Column(String)
    status = Column(Enum(JobStatus), default=JobStatus.INCOMPLETE)
    start_date = Column(DateTime, nullable=False)
    end_date = Column(DateTime, nullable=False)
    created_date = Column(DateTime, default=utc_now)
    data_sets_int = Column(Integer, nullable=False)
    season_id = Column(Integer, ForeignKey("season.id"))

    errors = relationship("ScrapeError", backref="job")

    @hybrid_property
    def bbref_games_for_date(self):
        return self.data_sets_int & DataSet.BBREF_GAMES_FOR_DATE == DataSet.BBREF_GAMES_FOR_DATE

    @hybrid_property
    def bbref_boxscores(self):
        return self.data_sets_int & DataSet.BBREF_BOXSCORES == DataSet.BBREF_BOXSCORES

    @hybrid_property
    def brooks_games_for_date(self):
        return self.data_sets_int & DataSet.BROOKS_GAMES_FOR_DATE == DataSet.BROOKS_GAMES_FOR_DATE

    @hybrid_property
    def brooks_pitch_logs(self):
        return self.data_sets_int & DataSet.BROOKS_PITCH_LOGS == DataSet.BROOKS_PITCH_LOGS

    @hybrid_property
    def brooks_pitchfx(self):
        return self.data_sets_int & DataSet.BROOKS_PITCHFX == DataSet.BROOKS_PITCHFX

    @hybrid_property
    def data_sets(self):
        data_sets = []
        if self.brooks_games_for_date:
            data_sets.append(DataSet.BROOKS_GAMES_FOR_DATE)
        if self.brooks_pitch_logs:
            data_sets.append(DataSet.BROOKS_PITCH_LOGS)
        if self.brooks_pitchfx:
            data_sets.append(DataSet.BROOKS_PITCHFX)
        if self.bbref_games_for_date:
            data_sets.append(DataSet.BBREF_GAMES_FOR_DATE)
        if self.bbref_boxscores:
            data_sets.append(DataSet.BBREF_BOXSCORES)
        return sorted(data_sets)

    @hybrid_property
    def created_date_str(self):
        created_date_utc = make_tzaware(self.created_date, use_tz=timezone.utc, localize=False)
        return localized_dt_string(created_date_utc, use_tz=get_local_utcoffset())

    @hybrid_property
    def url_set_filepath(self):
        return NODEJS_INBOX.joinpath(f"{self.id}.json")

    @hybrid_property
    def scraped_html_root_folder(self):
        return NODEJS_OUTBOX.joinpath(self.id)

    @hybrid_property
    def scraped_html_folders(self):
        folder_dict = {
            data_set: self.scraped_html_root_folder.joinpath(data_set.name.lower()) for data_set in self.data_sets
        }
        for folderpath in folder_dict.values():
            folderpath.mkdir(parents=True, exist_ok=True)
        return folder_dict

    @hybrid_property
    def date_range(self):
        return get_date_range(self.start_date, self.end_date)

    @hybrid_property
    def total_days(self):
        return len(self.date_range)

    @hybrid_property
    def job_details(self):
        job_name = f"{self.name} (ID: {self.id})" if self.name != self.id else self.id
        return {
            "Job Name": job_name,
            "Status": self.status.name,
            "MLB Season": self.season.name,
            "Start Date": self.start_date.strftime(DATE_ONLY),
            "End Date": self.end_date.strftime(DATE_ONLY),
            "Created At": self.created_date_str,
            "Data Sets": "\n\t      ".join(DATA_SET_TO_NAME_MAP[ds] for ds in self.data_sets),
        }

    @hybrid_property
    def error_messages(self):
        return "\n".join(str(error) for error in self.errors)

    def __repr__(self):
        return f"<ScrapeJob id={self.id}, status={self.status}>"

    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}

    @classmethod
    def from_user_params(cls, db_session, data_sets, start_date, end_date, season, job_name):
        new_job_dict = {
            "data_sets_int": sum(int(ds) for ds in data_sets),
            "start_date": start_date,
            "end_date": end_date,
            "season_id": season.id,
        }
        new_scrape_job = cls(**new_job_dict)
        db_session.add(new_scrape_job)
        db_session.commit()
        new_scrape_job.name = job_name if not string_is_null_or_blank(job_name) else str(new_scrape_job.id)
        db_session.commit()
        return new_scrape_job

    @classmethod
    def find_by_id(cls, db_session, job_id):
        return db_session.query(cls).filter_by(id=job_id).first()

    @classmethod
    def get_all_active_jobs(cls, db_session):
        return [job for job in db_session.query(cls).all() if job.status == JobStatus.INCOMPLETE]

    @classmethod
    def get_all_failed_jobs(cls, db_session):
        return [job for job in db_session.query(cls).all() if job.status == JobStatus.ERROR]

    @classmethod
    def get_all_complete_jobs(cls, db_session):
        return [job for job in db_session.query(cls).all() if job.status == JobStatus.COMPLETE]

    @classmethod
    def get_all_jobs_grouped_sorted(cls, db_session):
        all_jobs = db_session.query(cls).all()
        return group_and_sort_list(all_jobs, "status", "created_date", sort_all_desc=True)



from vigorish.setup.create_relationships import create_relationships
from vigorish.setup.populate_base_tables import populate_base_tables
from vigorish.setup.populate_pitch_type_perc import populate_pitch_type_perc
from vigorish.setup.populate_players import populate_players
from vigorish.setup.populate_seasons import populate_seasons
from vigorish.setup.populate_status_tables import populate_status_tables
from vigorish.setup.populate_teams import populate_teams


def populate_tables(app, csv_folder, json_folder):
    db_session = app.db_session
    result = (
        populate_base_tables(db_session)
        .on_success(populate_seasons, db_session)
        .on_success(populate_status_tables, db_session)
        .on_success(populate_teams, db_session, csv_folder)
        .on_success(populate_players, app, csv_folder)
        .on_success(populate_pitch_type_perc, app, json_folder)
        .on_success(create_relationships, db_session)
    )
    if result.failure:
        db_session.rollback()
        return result
    db_session.commit()
    return result


def populate_tables_for_restore(app, csv_folder, json_folder):
    db_session = app.db_session
    result = (
        populate_base_tables(db_session)
        .on_success(populate_seasons, db_session)
        .on_success(populate_teams, db_session, csv_folder)
        .on_success(populate_players, app, csv_folder)
        .on_success(populate_pitch_type_perc, app, json_folder)
        .on_success(create_relationships, db_session)
    )
    if result.failure:
        db_session.rollback()
        return result
    db_session.commit()
    return result



import vigorish.database as db
from vigorish.enums import DataSet
from vigorish.util.result import Result


def update_status_bbref_boxscore_list(scraped_data, db_session, new_bbref_game_ids, apply_patch_list=True):
    for bbref_game_id in new_bbref_game_ids:
        boxscore = scraped_data.get_bbref_boxscore(bbref_game_id, apply_patch_list)
        if not boxscore:
            error = f"Failed to retrieve {DataSet.BBREF_BOXSCORES} (URL ID: {bbref_game_id})"
            return Result.Fail(error)
        result = update_status_bbref_boxscore(db_session, boxscore)
        if result.failure:
            return result
    return Result.Ok()


def update_status_bbref_boxscore(db_session, boxscore):
    result = update_game_status_records(db_session, boxscore)
    if result.failure:
        return result
    db_session.commit()
    return Result.Ok()


def update_game_status_records(db_session, boxscore):
    try:
        game_status = db.GameScrapeStatus.find_by_bbref_game_id(db_session, boxscore.bbref_game_id)
        if not game_status:
            error = f"scrape_status_game does not contain an entry for bbref_game_id: {boxscore.bbref_game_id}"
            return Result.Fail(error)
        game_status.scraped_bbref_boxscore = 1
        game_status.pitch_app_count_bbref = boxscore.pitch_appearance_count
        game_status.total_pitch_count_bbref = boxscore.pitch_count
        return Result.Ok()
    except Exception as e:
        return Result.Fail(f"Error: {repr(e)}")



from sqlalchemy import func, join, select
from sqlalchemy_utils import create_view

import vigorish.database as db


class Season_Game_View(db.Base):
    __table__ = create_view(
        name="season_game_status",
        selectable=select(
            [
                db.Season.id.label("id"),
                func.count(db.GameScrapeStatus.id).label("total_games"),
                func.sum(db.GameScrapeStatus.scraped_bbref_boxscore).label("total_scraped_bbref_boxscore"),
                func.sum(db.GameScrapeStatus.scraped_brooks_pitch_logs).label("total_scraped_brooks_pitch_logs"),
                func.sum(db.GameScrapeStatus.combined_data_success).label("total_combined_data_success"),
                func.sum(db.GameScrapeStatus.combined_data_fail).label("total_combined_data_fail"),
                func.sum(db.GameScrapeStatus.pitch_app_count_bbref).label("total_pitch_app_count_bbref"),
                func.sum(db.GameScrapeStatus.pitch_app_count_brooks).label("total_pitch_app_count_brooks"),
                func.sum(db.GameScrapeStatus.total_pitch_count_bbref).label("total_pitch_count_bbref"),
            ]
        )
        .select_from(
            join(
                db.Season,
                db.GameScrapeStatus,
                db.Season.id == db.GameScrapeStatus.season_id,
            )
        )
        .group_by(db.Season.id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )



import vigorish.database as db
from vigorish.enums import DataSet, ScrapeCondition
from vigorish.scrape.bbref_boxscores.parse_html import parse_bbref_boxscore
from vigorish.scrape.scrape_task import ScrapeTaskABC
from vigorish.status.update_status_bbref_boxscores import update_status_bbref_boxscore
from vigorish.util.dt_format_strings import DATE_ONLY
from vigorish.util.result import Result


class ScrapeBBRefBoxscores(ScrapeTaskABC):
    def __init__(self, app, db_job):
        self.data_set = DataSet.BBREF_BOXSCORES
        super().__init__(app, db_job)

    def check_prerequisites(self, game_date):
        bbref_games_for_date = db.DateScrapeStatus.verify_bbref_daily_dashboard_scraped_for_date(
            self.db_session, game_date
        )
        if bbref_games_for_date:
            return Result.Ok()
        date_str = game_date.strftime(DATE_ONLY)
        error = (
            f"BBref games for date {date_str} have not been scraped, unable to scrape BBref "
            "boxscores until this has been done."
        )
        return Result.Fail(error)

    def check_current_status(self, game_date):
        if self.scrape_condition == ScrapeCondition.ALWAYS:
            return Result.Ok()
        scraped_bbref_boxscores = db.DateScrapeStatus.verify_all_bbref_boxscores_scraped_for_date(
            self.db_session, game_date
        )
        return Result.Ok() if not scraped_bbref_boxscores else Result.Fail("skip")

    def parse_html(self, url_details):
        return parse_bbref_boxscore(url_details.html, url_details.url, url_details.url_id)

    def update_status(self, parsed_data):
        return update_status_bbref_boxscore(self.db_session, parsed_data)



"""Pitchfx measurements for a single pitch scraped from brooksbaseball.com."""
from dataclasses import dataclass, field
from datetime import datetime, timezone

from vigorish.enums import PitchType
from vigorish.util.datetime_util import TIME_ZONE_NEW_YORK
from vigorish.util.dt_format_strings import DT_AWARE
from vigorish.util.regex import PFX_TIMESTAMP_REGEX
from vigorish.util.string_helpers import validate_bbref_game_id


@dataclass
class BrooksPitchFxData:
    """Pitchfx measurements for a single pitch scraped from brooksbaseball.com."""

    pitcher_name: str = field(repr=False, default="")
    pitch_app_id: str = ""
    pitcher_id: int = field(repr=False, default=0)
    batter_id: int = field(repr=False, default=0)
    pitcher_team_id_bb: str = field(repr=False, default="")
    opponent_team_id_bb: str = field(repr=False, default="")
    bb_game_id: str = field(repr=False, default="")
    bbref_game_id: str = field(repr=False, default="")
    table_row_number: int = field(repr=False, default=0)
    park_sv_id: str = field(repr=False, default="")
    play_guid: str = field(repr=False, default="")
    ab_total: int = field(repr=False, default=0)
    ab_count: int = field(repr=False, default=0)
    ab_id: int = field(repr=False, default=0)
    des: str = field(repr=False, default="")
    type: str = field(repr=False, default="")
    id: int = 0
    sz_top: float = field(repr=False, default=0.0)
    sz_bot: float = field(repr=False, default=0.0)
    pfx_xdatafile: float = field(repr=False, default=0.0)
    pfx_zdatafile: float = field(repr=False, default=0.0)
    mlbam_pitch_name: str = ""
    zone_location: int = field(repr=False, default=0)
    pitch_con: float = field(repr=False, default=0.0)
    stand: str = field(repr=False, default="")
    strikes: int = field(repr=False, default=0)
    balls: int = field(repr=False, default=0)
    p_throws: str = field(repr=False, default="")
    pdes: str = field(repr=False, default="")
    spin: float = field(repr=False, default=0.0)
    norm_ht: float = field(repr=False, default=0.0)
    inning: int = field(repr=False, default=0)
    tstart: float = field(repr=False, default=0.0)
    vystart: float = field(repr=False, default=0.0)
    ftime: float = field(repr=False, default=0.0)
    pfx_x: float = field(repr=False, default=0.0)
    pfx_z: float = field(repr=False, default=0.0)
    uncorrected_pfx_x: float = field(repr=False, default=0.0)
    uncorrected_pfx_z: float = field(repr=False, default=0.0)
    x0: float = field(repr=False, default=0.0)
    y0: float = field(repr=False, default=0.0)
    z0: float = field(repr=False, default=0.0)
    vx0: float = field(repr=False, default=0.0)
    vy0: float = field(repr=False, default=0.0)
    vz0: float = field(repr=False, default=0.0)
    ax: float = field(repr=False, default=0.0)
    ay: float = field(repr=False, default=0.0)
    az: float = field(repr=False, default=0.0)
    start_speed: float = field(repr=False, default=0.0)
    px: float = field(repr=False, default=0.0)
    pz: float = field(repr=False, default=0.0)
    pxold: float = field(repr=False, default=0.0)
    pzold: float = field(repr=False, default=0.0)
    tm_spin: int = field(repr=False, default=0)
    sb: int = field(repr=False, default=0)
    game_start_time_str: str = field(repr=False, default="")
    time_pitch_thrown_str: str = field(repr=False, default="")
    has_zone_location: bool = field(repr=False, default=False)
    batter_did_swing: bool = field(repr=False, default=False)
    batter_made_contact: bool = field(repr=False, default=False)
    called_strike: bool = field(repr=False, default=False)
    swinging_strike: bool = field(repr=False, default=False)
    inside_strike_zone: bool = field(repr=False, default=False)
    outside_strike_zone: bool = field(repr=False, default=False)
    swing_inside_zone: bool = field(repr=False, default=False)
    swing_outside_zone: bool = field(repr=False, default=False)
    contact_inside_zone: bool = field(repr=False, default=False)
    contact_outside_zone: bool = field(repr=False, default=False)
    is_batted_ball: bool = field(repr=False, default=False)
    is_ground_ball: bool = field(repr=False, default=False)
    is_fly_ball: bool = field(repr=False, default=False)
    is_line_drive: bool = field(repr=False, default=False)
    is_pop_up: bool = field(repr=False, default=False)
    is_final_pitch_of_ab: bool = field(repr=False, default=False)
    ab_result_out: bool = field(repr=False, default=False)
    ab_result_hit: bool = field(repr=False, default=False)
    ab_result_single: bool = field(repr=False, default=False)
    ab_result_double: bool = field(repr=False, default=False)
    ab_result_triple: bool = field(repr=False, default=False)
    ab_result_homerun: bool = field(repr=False, default=False)
    ab_result_bb: bool = field(repr=False, default=False)
    ab_result_ibb: bool = field(repr=False, default=False)
    ab_result_k: bool = field(repr=False, default=False)
    ab_result_hbp: bool = field(repr=False, default=False)
    ab_result_error: bool = field(repr=False, default=False)
    ab_result_sac_hit: bool = field(repr=False, default=False)
    ab_result_sac_fly: bool = field(repr=False, default=False)
    ab_result_unclear: bool = field(repr=False, default=False)
    pbp_play_result: str = field(repr=False, default="")
    pbp_runs_outs_result: str = field(repr=False, default="")
    is_patched: bool = field(repr=False, default=False)
    is_invalid_ibb: bool = field(repr=False, default=False)
    is_out_of_sequence: bool = field(repr=False, default=False)

    @property
    def game_start_time(self):
        return datetime.strptime(self.game_start_time_str, DT_AWARE) if self.game_start_time_str else None

    @property
    def time_pitch_thrown(self):
        match = PFX_TIMESTAMP_REGEX.match(self.park_sv_id)
        if not match:
            return None
        group_dict = match.groupdict()
        if int(group_dict["second"]) >= 60:
            group_dict["second"] = 0
        game_dict = validate_bbref_game_id(self.bbref_game_id).value
        timestamp = datetime(
            game_dict["game_date"].year,
            int(group_dict["month"]),
            int(group_dict["day"]),
            int(group_dict["hour"]),
            int(group_dict["minute"]),
            int(group_dict["second"]),
        )
        thrown_replace_tz = timestamp.replace(tzinfo=TIME_ZONE_NEW_YORK)
        thrown_as_tz = timestamp.replace(tzinfo=timezone.utc).astimezone(TIME_ZONE_NEW_YORK)
        if not self.game_start_time:
            return thrown_as_tz
        seconds_since_thrown_replace_tz = (thrown_replace_tz - self.game_start_time).total_seconds()
        seconds_since_thrown_as_tz = (thrown_as_tz - self.game_start_time).total_seconds()
        if seconds_since_thrown_as_tz < 0 and seconds_since_thrown_replace_tz < 0:
            return None
        if seconds_since_thrown_as_tz > 0 and seconds_since_thrown_replace_tz < 0:
            return thrown_as_tz
        if seconds_since_thrown_as_tz < 0 and seconds_since_thrown_replace_tz > 0:
            return thrown_replace_tz
        return thrown_as_tz if seconds_since_thrown_as_tz < seconds_since_thrown_replace_tz else thrown_replace_tz

    @property
    def seconds_since_game_start(self):
        if not self.time_pitch_thrown or not self.game_start_time:
            return 0
        return int((self.time_pitch_thrown - self.game_start_time).total_seconds())

    def as_dict(self):
        """Convert pitch log to a dictionary."""
        return {
            "__brooks_pitchfx_data__": True,
            "pitcher_name": self.pitcher_name,
            "pitch_app_id": self.pitch_app_id,
            "pitcher_id": self.pitcher_id,
            "batter_id": self.batter_id,
            "pitcher_team_id_bb": self.pitcher_team_id_bb,
            "opponent_team_id_bb": self.opponent_team_id_bb,
            "bb_game_id": self.bb_game_id,
            "bbref_game_id": self.bbref_game_id,
            "table_row_number": self.table_row_number,
            "park_sv_id": self.park_sv_id,
            "play_guid": self.play_guid,
            "ab_total": self.ab_total,
            "ab_count": self.ab_count,
            "ab_id": self.ab_id,
            "des": self.des,
            "type": self.type,
            "id": self.id,
            "sz_top": self.sz_top,
            "sz_bot": self.sz_bot,
            "pfx_xdatafile": self.pfx_xdatafile,
            "pfx_zdatafile": self.pfx_zdatafile,
            "mlbam_pitch_name": self.mlbam_pitch_name,
            "zone_location": self.zone_location,
            "pitch_con": self.pitch_con,
            "stand": self.stand,
            "strikes": self.strikes,
            "balls": self.balls,
            "p_throws": self.p_throws,
            "pdes": self.pdes,
            "spin": self.spin,
            "norm_ht": self.norm_ht,
            "inning": self.inning,
            "tstart": self.tstart,
            "vystart": self.vystart,
            "ftime": self.ftime,
            "pfx_x": self.pfx_x,
            "pfx_z": self.pfx_z,
            "uncorrected_pfx_x": self.uncorrected_pfx_x,
            "uncorrected_pfx_z": self.uncorrected_pfx_z,
            "x0": self.x0,
            "y0": self.y0,
            "z0": self.z0,
            "vx0": self.vx0,
            "vy0": self.vy0,
            "vz0": self.vz0,
            "ax": self.ax,
            "ay": self.ay,
            "az": self.az,
            "start_speed": self.start_speed,
            "px": self.px,
            "pz": self.pz,
            "pxold": self.pxold,
            "pzold": self.pzold,
            "tm_spin": self.tm_spin,
            "sb": self.sb,
            "game_start_time_str": self.game_start_time_str,
            "time_pitch_thrown_str": self.get_time_pitch_thrown_str(),
            "has_zone_location": self.has_zone_location,
            "batter_did_swing": self.batter_did_swing,
            "batter_made_contact": self.batter_made_contact,
            "called_strike": self.called_strike,
            "swinging_strike": self.swinging_strike,
            "inside_strike_zone": self.inside_strike_zone,
            "outside_strike_zone": self.outside_strike_zone,
            "swing_inside_zone": self.swing_inside_zone,
            "swing_outside_zone": self.swing_outside_zone,
            "contact_inside_zone": self.contact_inside_zone,
            "contact_outside_zone": self.contact_outside_zone,
            "is_batted_ball": self.is_batted_ball,
            "is_ground_ball": self.is_ground_ball,
            "is_fly_ball": self.is_fly_ball,
            "is_line_drive": self.is_line_drive,
            "is_pop_up": self.is_pop_up,
            "is_final_pitch_of_ab": self.is_final_pitch_of_ab,
            "ab_result_out": self.is_final_pitch_of_ab,
            "ab_result_hit": self.is_final_pitch_of_ab,
            "ab_result_single": self.is_final_pitch_of_ab,
            "ab_result_double": self.is_final_pitch_of_ab,
            "ab_result_triple": self.is_final_pitch_of_ab,
            "ab_result_homerun": self.is_final_pitch_of_ab,
            "ab_result_bb": self.is_final_pitch_of_ab,
            "ab_result_ibb": self.is_final_pitch_of_ab,
            "ab_result_k": self.is_final_pitch_of_ab,
            "ab_result_hbp": self.is_final_pitch_of_ab,
            "ab_result_error": self.ab_result_error,
            "ab_result_sac_hit": self.is_final_pitch_of_ab,
            "ab_result_sac_fly": self.is_final_pitch_of_ab,
            "ab_result_unclear": self.is_final_pitch_of_ab,
            "pitch_type_int": int(PitchType.from_abbrev(self.mlbam_pitch_name)),
            "pbp_play_result": self.pbp_play_result,
            "pbp_runs_outs_result": self.pbp_runs_outs_result,
            "is_patched": self.is_patched,
            "is_invalid_ibb": self.is_invalid_ibb,
            "is_out_of_sequence": self.is_out_of_sequence,
        }

    def get_time_pitch_thrown_str(self):
        return self.time_pitch_thrown.strftime(DT_AWARE) if self.time_pitch_thrown else ""



from functools import cached_property
from typing import Dict, List

from sqlalchemy.engine import Engine
from sqlalchemy.orm import Session

import vigorish.database as db
from vigorish.app import Vigorish
from vigorish.data.metrics import BatStatsMetrics, PitchStatsMetrics
from vigorish.data.scraped_data import ScrapedData
from vigorish.enums import DefensePosition


class TeamData:
    def __init__(self, app: Vigorish, team_id_bbref: str, year: int):
        self.app = app
        self.team_id_bbref = team_id_bbref
        self.year = year
        self.db_engine: Engine = self.app.db_engine
        self.db_session: Session = self.app.db_session
        self.scraped_data: ScrapedData = self.app.scraped_data
        self.team: db.Team = db.Team.find_by_team_id_and_year(self.db_session, self.team_id_bbref, self.year)

    @cached_property
    def pitch_stats(self) -> PitchStatsMetrics:
        return self.scraped_data.get_pitch_stats_for_team(self.team_id_bbref, self.year)

    @cached_property
    def pitch_stats_by_year(self) -> Dict[int, PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_by_year_for_team(self.team_id_bbref)

    @cached_property
    def pitch_stats_by_player(self) -> List[PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_by_player_for_team(self.team_id_bbref, self.year)

    @cached_property
    def pitch_stats_for_sp(self) -> PitchStatsMetrics:
        return self.scraped_data.get_pitch_stats_for_sp_for_team(self.team_id_bbref, self.year)

    @cached_property
    def pitch_stats_for_sp_by_year(self) -> Dict[int, PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_for_sp_by_year_for_team(self.team_id_bbref)

    @cached_property
    def pitch_stats_for_sp_by_player(self) -> List[PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_for_sp_by_player_for_team(self.team_id_bbref, self.year)

    @cached_property
    def pitch_stats_for_rp(self) -> PitchStatsMetrics:
        return self.scraped_data.get_pitch_stats_for_rp_for_team(self.team_id_bbref, self.year)

    @cached_property
    def pitch_stats_for_rp_by_year(self) -> Dict[int, PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_for_rp_by_year_for_team(self.team_id_bbref)

    @cached_property
    def pitch_stats_for_rp_by_player(self) -> List[PitchStatsMetrics]:
        return self.scraped_data.get_pitch_stats_for_rp_by_player_for_team(self.team_id_bbref, self.year)

    @cached_property
    def bat_stats(self) -> BatStatsMetrics:
        return self.scraped_data.get_bat_stats_for_team(self.team_id_bbref, self.year)

    @cached_property
    def bat_stats_by_year(self) -> Dict[int, BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_by_year_for_team(self.team_id_bbref)

    @cached_property
    def bat_stats_by_player(self) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_by_player_for_team(self.team_id_bbref, self.year)

    @cached_property
    def bat_stats_by_lineup_spot(self) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_by_lineup_spot_for_team(self.team_id_bbref, self.year)

    @cached_property
    def bat_stats_by_defpos(self) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_by_defpos_for_team(self.team_id_bbref, self.year)

    @cached_property
    def bat_stats_for_starters(self) -> BatStatsMetrics:
        return self.scraped_data.get_bat_stats_for_starters_for_team(self.team_id_bbref, self.year)

    @cached_property
    def bat_stats_for_starters_by_year(self) -> Dict[int, BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_for_starters_by_year_for_team(self.team_id_bbref)

    @cached_property
    def bat_stats_for_starters_by_player(self) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_for_starters_by_player_for_team(self.team_id_bbref, self.year)

    @cached_property
    def bat_stats_for_subs(self) -> BatStatsMetrics:
        return self.scraped_data.get_bat_stats_for_subs_for_team(self.team_id_bbref, self.year)

    @cached_property
    def bat_stats_for_subs_by_year(self) -> Dict[int, BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_for_subs_by_year_for_team(self.team_id_bbref)

    @cached_property
    def bat_stats_for_subs_by_player(self) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_for_subs_by_player_for_team(self.team_id_bbref, self.year)

    def get_bat_stats_for_lineup_spot_by_year(self, lineup_spot: int) -> Dict[int, BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_for_lineup_spot_by_year_for_team(lineup_spot, self.team_id_bbref)

    def get_bat_stats_for_lineup_spot_by_player(self, lineup_spot: int) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_for_lineup_spot_by_player_for_team(
            lineup_spot, self.team_id_bbref, self.year
        )

    def get_bat_stats_for_defpos_by_year(self, def_position: DefensePosition) -> Dict[int, BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_for_defpos_by_year_for_team(def_position, self.team_id_bbref)

    def get_bat_stats_for_defpos_by_player(self, def_position: DefensePosition) -> List[BatStatsMetrics]:
        return self.scraped_data.get_bat_stats_for_defpos_by_player_for_team(
            def_position, self.team_id_bbref, self.year
        )



from dataclasses import dataclass


@dataclass
class BatchJobSettings:
    batched_scraping_enabled: bool
    batch_size_is_random: bool
    batch_size_uniform: int
    batch_size_random_min: int
    batch_size_random_max: int

    def __str__(self):
        if not self.batched_scraping_enabled:
            return "Batched scraping is not enabled"
        if not self.batch_size_is_random:
            return f"Batch size is uniform ({self.batch_size_uniform} URLs)"
        return f"Batch size is random ({self.batch_size_random_min}-{self.batch_size_random_max} URLs)"

    def to_dict(self):
        return {
            "CREATE_BATCHED_SCRAPE_JOBS": self.batched_scraping_enabled,
            "USE_IRREGULAR_BATCH_SIZES": self.batch_size_is_random,
            "BATCH_SIZE": self.batch_size_uniform,
            "BATCH_SIZE_MIN": self.batch_size_random_min,
            "BATCH_SIZE_MAX": self.batch_size_random_max,
        }

    @staticmethod
    def from_config(config_dict):
        batched_scraping_enabled = config_dict.get("CREATE_BATCHED_SCRAPE_JOBS", True)
        batch_size_is_random = config_dict.get("USE_IRREGULAR_BATCH_SIZES", True)
        batch_size_uniform = config_dict.get("BATCH_SIZE", True)
        batch_size_random_min = config_dict.get("BATCH_SIZE_MIN", True)
        batch_size_random_max = config_dict.get("BATCH_SIZE_MAX", True)
        return BatchJobSettings(
            batched_scraping_enabled=batched_scraping_enabled,
            batch_size_is_random=batch_size_is_random,
            batch_size_uniform=batch_size_uniform,
            batch_size_random_min=batch_size_random_min,
            batch_size_random_max=batch_size_random_max,
        )

    @staticmethod
    def null_object():
        return {
            "CREATE_BATCHED_SCRAPE_JOBS": None,
            "USE_IRREGULAR_BATCH_SIZES": None,
            "BATCH_SIZE": None,
            "BATCH_SIZE_MIN": None,
            "BATCH_SIZE_MAX": None,
        }



"""Menu that allows the user to view and modify all settings in vig.config.json."""
from vigorish.cli.menu import Menu
from vigorish.cli.menu_items.config_setting import ConfigSetting
from vigorish.cli.menu_items.return_to_parent import ReturnToParent
from vigorish.constants import EMOJIS


class ConfigSettingsMenu(Menu):
    def __init__(self, app):
        super().__init__(app)
        self.menu_text = "You can modify any setting in the list below:"
        self.menu_item_text = "Config File Settings"
        self.menu_item_emoji = EMOJIS.get("GEAR", "")

    def populate_menu_items(self):
        self.menu_items = [ConfigSetting(self.app, name) for name in self.config.all_settings.keys()]
        self.menu_items.append(ReturnToParent(self.app, "Return to Settings/Admin "))
        self.menu_items.insert(0, ReturnToParent(self.app, "Return to Settings/Admin"))



"""Helper functions that handle and produce numeric values."""
import math
from datetime import date

from vigorish.util.string_helpers import try_parse_int

ONE_KB = 1024
ONE_MB = ONE_KB * 1024
ONE_GB = ONE_MB * 1024
ONE_PERCENT = 0.01


def is_even(x):
    return x & 1 == 0


def validate_year_value(year):
    if not year:
        error = 'No value provided for "year" parameter, unable to resolve folder path.'
        raise ValueError(error)
    if isinstance(year, str):
        year_str = year
        year = try_parse_int(year_str)
        if not year:
            raise ValueError(f'Failed to parse int value from string "{year_str}"')
    if isinstance(year, bool) or not isinstance(year, int):
        raise TypeError(f'"year" parameter must be int value (not "{type(year)}").')
    if year < 1900:
        raise ValueError(f"Data is not collected for year={year}")
    if year > date.today().year:
        raise ValueError(f'"{year}" is not valid since it is a future year')
    return year


def trim_data_set(samples, st_dev_limit=2):
    if not samples or not isinstance(samples, list):
        raise ValueError("Input list is either empty or invalid")
    mean = sum(samples) / len(samples)
    st_dev = get_standard_deviation(samples)
    return [x for x in samples if x > (mean - st_dev_limit * st_dev) and x < (mean + st_dev_limit * st_dev)]


def get_standard_deviation(samples):
    if not samples or not isinstance(samples, list):
        raise ValueError("Input list is either empty or invalid")
    mean = sum(samples) / len(samples)
    summed_squares = sum(math.pow((sample - mean), 2) for sample in samples)
    return math.sqrt(summed_squares / (len(samples) - 1))



from pathlib import Path

APP_FOLDER = Path(__file__).parent.parent
ROOT_FOLDER = APP_FOLDER.parent.parent
VIG_FOLDER = Path.home().joinpath(".vig")
DOTENV_FILE = VIG_FOLDER.joinpath(".env")
CONFIG_FILE = VIG_FOLDER.joinpath("vig.config.json")

SQLITE_DEV_URL = f"sqlite:///{VIG_FOLDER.joinpath('vig_dev.db')}"
SQLITE_PROD_URL = f"sqlite:///{VIG_FOLDER.joinpath('vig.db')}"

NIGHTMAREJS_FOLDER = APP_FOLDER.joinpath("nightmarejs")
NODEJS_INBOX = NIGHTMAREJS_FOLDER.joinpath("inbox")
NODEJS_OUTBOX = NIGHTMAREJS_FOLDER.joinpath("outbox")
NODEJS_SCRIPT = NIGHTMAREJS_FOLDER.joinpath("scrape_job.js")

JSON_FOLDER = APP_FOLDER.joinpath("setup/json")
CSV_FOLDER = APP_FOLDER.joinpath("setup/csv")
PEOPLE_CSV = CSV_FOLDER.joinpath("People.csv")
TEAM_CSV = CSV_FOLDER.joinpath("Teams.csv")
PLAYER_ID_MAP_CSV = CSV_FOLDER.joinpath("bbref_player_id_map.csv")
PLAYER_TEAM_MAP_CSV = CSV_FOLDER.joinpath("bbref_player_team_map.csv")



from dataclasses import dataclass, field
from typing import List

from vigorish.scrape.bbref_boxscores.models.pbp_event import BBRefPlayByPlayEvent
from vigorish.scrape.bbref_boxscores.models.pbp_other import BBRefPlayByPlayMiscEvent
from vigorish.scrape.bbref_boxscores.models.pbp_substitution import BBRefInGameSubstitution
from vigorish.util.list_helpers import as_dict_list


@dataclass
class BBRefHalfInning:
    inning_id: str = ""
    inning_label: str = ""
    begin_inning_summary: str = ""
    end_inning_summary: str = ""
    inning_total_runs: str = "0"
    inning_total_hits: str = "0"
    inning_total_errors: str = "0"
    inning_total_left_on_base: str = "0"
    away_team_runs_after_inning: str = "0"
    home_team_runs_after_inning: str = "0"
    game_events: List[BBRefPlayByPlayEvent] = field(default_factory=list)
    substitutions: List[BBRefInGameSubstitution] = field(default_factory=list)
    misc_events: List[BBRefPlayByPlayMiscEvent] = field(default_factory=list)

    def as_dict(self):
        """Convert half-inning to a dictionary."""
        return {
            "__bbref_boxscore_half_inning__": True,
            "inning_id": self.inning_id,
            "inning_label": self.inning_label,
            "begin_inning_summary": self.begin_inning_summary,
            "end_inning_summary": self.end_inning_summary,
            "inning_total_runs": int(self.inning_total_runs),
            "inning_total_hits": int(self.inning_total_hits),
            "inning_total_errors": int(self.inning_total_errors),
            "inning_total_left_on_base": int(self.inning_total_left_on_base),
            "away_team_runs_after_inning": int(self.away_team_runs_after_inning),
            "home_team_runs_after_inning": int(self.home_team_runs_after_inning),
            "game_events": as_dict_list(self.game_events),
            "substitutions": as_dict_list(self.substitutions),
            "misc_events": as_dict_list(self.misc_events),
        }



from dataclasses import dataclass

from vigorish.cli.components.util import print_heading, print_message


@dataclass(frozen=True)
class DisplayPage:
    text: str
    heading: str = None

    def display(self, heading_color=None, text_color=None, wrap_text=True):
        if self.heading:
            print_heading(self.heading, fg=heading_color)
        for s in self.text:
            print_message(s, fg=text_color, wrap=wrap_text)



from tqdm import tqdm

import vigorish.database as db
from vigorish.util.result import Result


def populate_status_tables(db_session):
    try:
        mlb_seasons = db.Season.get_all_regular_seasons(db_session)
        with tqdm(
            total=len(mlb_seasons),
            unit="season",
            mininterval=0.12,
            maxinterval=5,
            position=0,
            ncols=90,
        ) as pbar:
            for season in mlb_seasons:
                pbar.set_description("Populating status_date table...")
                for game_date in season.get_date_range():
                    scrape_status = db.DateScrapeStatus(game_date=game_date, season_id=season.id)
                    db_session.add(scrape_status)
                pbar.update()
        db_session.commit()
        return Result.Ok()
    except Exception as e:
        error = f"Error: {repr(e)}"
        db_session.rollback()
        return Result.Fail(error)



from sqlalchemy import Column, DateTime, Float, Integer

import vigorish.database as db
from vigorish.util.datetime_util import utc_now


class TimeBetweenPitches(db.Base):

    __tablename__ = "time_between_pitches"
    id = Column(Integer, primary_key=True)
    between_pitches_sample_size = Column(Integer, default=0)
    between_pitches_total_sec = Column(Integer, default=0)
    between_pitches_avg_sec = Column(Float, default=0.0)
    between_pitches_min_sec = Column(Float, default=0.0)
    between_pitches_max_sec = Column(Float, default=0.0)
    between_at_bats_sample_size = Column(Integer, default=0)
    between_at_bats_total_sec = Column(Integer, default=0)
    between_at_bats_avg_sec = Column(Float, default=0.0)
    between_at_bats_min_sec = Column(Float, default=0.0)
    between_at_bats_max_sec = Column(Float, default=0.0)
    between_innings_sample_size = Column(Integer, default=0)
    between_innings_total_sec = Column(Integer, default=0)
    between_innings_avg_sec = Column(Float, default=0.0)
    between_innings_min_sec = Column(Float, default=0.0)
    between_innings_max_sec = Column(Float, default=0.0)
    calc_timestamp = Column(DateTime, default=utc_now)

    def as_dict(self):
        return {
            "time_between_pitches": {
                "total": self.between_pitches_sample_size,
                "count": self.between_pitches_total_sec,
                "avg": self.between_pitches_avg_sec,
                "max": self.between_pitches_max_sec,
                "min": self.between_pitches_min_sec,
                "range": self.between_pitches_max_sec - self.between_pitches_min_sec,
            },
            "time_between_at_bats": {
                "total": self.between_at_bats_sample_size,
                "count": self.between_at_bats_total_sec,
                "avg": self.between_at_bats_avg_sec,
                "max": self.between_at_bats_max_sec,
                "min": self.between_at_bats_min_sec,
                "range": self.between_at_bats_max_sec - self.between_at_bats_min_sec,
            },
            "time_between_innings": {
                "total": self.between_innings_sample_size,
                "count": self.between_innings_total_sec,
                "avg": self.between_innings_avg_sec,
                "max": self.between_innings_max_sec,
                "min": self.between_innings_min_sec,
                "range": self.between_innings_max_sec - self.between_innings_min_sec,
            },
            "timestamp": self.calc_timestamp,
        }

    @classmethod
    def from_calc_results(cls, db_session, calc_results):
        between_pitches = calc_results["time_between_pitches"]
        between_at_bats = calc_results["time_between_at_bats"]
        between_innings = calc_results["time_between_innings"]
        add_results = {
            "between_pitches_sample_size": between_pitches["total"],
            "between_pitches_total_sec": between_pitches["count"],
            "between_pitches_avg_sec": between_pitches["avg"],
            "between_pitches_min_sec": between_pitches["min"],
            "between_pitches_max_sec": between_pitches["max"],
            "between_at_bats_sample_size": between_at_bats["total"],
            "between_at_bats_total_sec": between_at_bats["count"],
            "between_at_bats_avg_sec": between_at_bats["avg"],
            "between_at_bats_min_sec": between_at_bats["min"],
            "between_at_bats_max_sec": between_at_bats["max"],
            "between_innings_sample_size": between_innings["total"],
            "between_innings_total_sec": between_innings["count"],
            "between_innings_avg_sec": between_innings["avg"],
            "between_innings_min_sec": between_innings["min"],
            "between_innings_max_sec": between_innings["max"],
        }
        time_between_pitches = cls(**add_results)
        db_session.add(time_between_pitches)
        db_session.commit()
        return time_between_pitches

    @classmethod
    def get_latest_results(cls, db_session):
        results = db_session.query(cls).all()
        if not results:
            return None
        results.sort(key=lambda x: x.calc_timestamp, reverse=True)
        return results[0].as_dict()



import os
from pathlib import Path

from vigorish.config.project_paths import CONFIG_FILE, DOTENV_FILE, SQLITE_PROD_URL
from vigorish.constants import ENV_VAR_NAMES
from vigorish.util.result import Result


class DotEnvFile:
    def __init__(self, dotenv_filepath=None):
        self.dotenv_filepath = dotenv_filepath
        if not self.dotenv_filepath:
            self.dotenv_filepath = Path(os.environ.get("DOTENV_FILE", DOTENV_FILE))
        self.read_dotenv_file()

    def read_dotenv_file(self):
        """Parse .env file to a dictionary of environment variables."""
        if not self.dotenv_filepath.exists():
            self.create_dotenv_file(
                config_file=os.environ.get("CONFIG_FILE", CONFIG_FILE),
                db_url=os.environ.get("DATABASE_URL", SQLITE_PROD_URL),
            )
        if not self.dotenv_filepath.is_file():
            raise TypeError(f"Unable to open file: {self.dotenv_filepath}")
        file_text = self.dotenv_filepath.read_text()
        env_var_split = [f for f in file_text.split("\n") if f and "=" in f]
        env_var_split = [s.split("=") for s in env_var_split]
        self.env_var_dict = {v[0]: v[1].strip('"').strip("'").strip() for v in env_var_split if len(v) == 2}
        if "CONFIG_FILE" not in self.env_var_dict:
            self.env_var_dict["CONFIG_FILE"] = CONFIG_FILE
        if "DATABASE_URL" not in self.env_var_dict:
            self.env_var_dict["DATABASE_URL"] = SQLITE_PROD_URL
        self.update_environment_variables()

    def create_dotenv_file(self, config_file, db_url):
        self.env_var_dict = {var_name: "" for var_name in ENV_VAR_NAMES}
        self.env_var_dict["CONFIG_FILE"] = config_file
        self.env_var_dict["DATABASE_URL"] = db_url
        self.write_dotenv_file()

    def update_environment_variables(self):
        for var_name, value in self.env_var_dict.items():
            os.environ[var_name] = value

    def get_current_value(self, env_var_name):
        if env_var_name not in ENV_VAR_NAMES:
            return None
        return self.env_var_dict.get(env_var_name, None)

    def change_value(self, env_var_name, new_value):
        if env_var_name not in ENV_VAR_NAMES:
            return Result.Fail(f"{env_var_name} is not a recognized environment variable.")
        self.env_var_dict[env_var_name] = new_value
        self.write_dotenv_file()
        self.read_dotenv_file()
        return Result.Ok()

    def restart_required_on_change(self, setting_name):
        return setting_name in ["CONFIG_FILE", "DATABASE_URL"]

    def write_dotenv_file(self):
        env_var_strings = [f"{name}={value}" for name, value in self.env_var_dict.items()]
        self.dotenv_filepath.write_text("\n".join(env_var_strings))



"""Menu item that allows the user to initialize/reset the database."""
import subprocess

from vigorish.cli.components import season_prompt, single_date_prompt, user_options_prompt
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.enums import StatusReport as Report
from vigorish.status.report_status import (
    report_date_range_status,
    report_season_status,
    report_status_single_date,
)
from vigorish.util.result import Result

PROMPT_TEXT = "Choose the type of report you wish to generate from the options below:"


class StatusReport(MenuItem):
    def __init__(self, app):
        # TODO: New Status Report menu option - Single Game. User can provide either BB or BR ID
        # TODO: Another Status Report option - Pitch Appearance. Simply call .display() method
        super().__init__(app)
        self.menu_item_text = "Status Reports"
        self.menu_item_emoji = EMOJIS.get("CHART")
        self.exit_menu = False

    def launch(self):
        subprocess.run(["clear"])
        result = self.report_options_prompt()
        if result.failure:
            return Result.Ok(self.exit_menu)
        report_type = result.value
        if report_type == "SEASON":
            result = self.season_status_report()
        if report_type == "SINGLE_DATE":
            result = self.single_date_report()
        if report_type == "DATE_RANGE":
            result = self.date_range_report()
        if result.failure:
            if "no report" in result.error:
                return Result.Ok(self.exit_menu)
            return result
        report_viewer = result.value
        report_viewer.launch()
        return Result.Ok(self.exit_menu)

    def report_options_prompt(self):
        choices = {
            f"{MENU_NUMBERS.get(1)}  Season": "SEASON",
            f"{MENU_NUMBERS.get(2)}  Single Date": "SINGLE_DATE",
            f"{MENU_NUMBERS.get(3)}  Date Range": "DATE_RANGE",
            f"{EMOJIS.get('BACK')} Return to Main Menu": None,
        }
        return user_options_prompt(choices, PROMPT_TEXT)

    def get_season_report_type_from_user(self):
        choice_text1 = f"{MENU_NUMBERS.get(1)}  Season Summary"
        choice_text2 = f"{MENU_NUMBERS.get(2)}  Dates Missing Data (Summary)"
        choice_text3 = f"{MENU_NUMBERS.get(3)}  All Dates In Season (Summary)"
        choice_text4 = f"{MENU_NUMBERS.get(4)}  Dates Missing Data (Detail)"
        choice_text5 = f"{MENU_NUMBERS.get(5)}  All Dates In Season (Detail)"
        choice_text6 = f"{MENU_NUMBERS.get(6)}  All Dates In Season + Missing PitchFx IDs (Detail)"
        choices = {
            choice_text1: Report.SEASON_SUMMARY,
            choice_text2: Report.DATE_SUMMARY_MISSING_DATA,
            choice_text3: Report.DATE_SUMMARY_ALL_DATES,
            choice_text4: Report.DATE_DETAIL_MISSING_DATA,
            choice_text5: Report.DATE_DETAIL_ALL_DATES,
            choice_text6: Report.DATE_DETAIL_MISSING_PITCHFX,
            f"{EMOJIS.get('BACK')} Return to Previous Menu": None,
        }
        return user_options_prompt(choices, PROMPT_TEXT)

    def get_single_date_report_type_from_user(self):
        choice_text1 = f"{MENU_NUMBERS.get(1)}  Detail Report"
        choice_text2 = f"{MENU_NUMBERS.get(2)}  Detail Report + Missing PitchFx IDs"
        choice_text3 = f"{MENU_NUMBERS.get(3)}  Detail Report + Missing PitchFx IDs + Game Status"
        choices = {
            choice_text1: Report.DATE_DETAIL_ALL_DATES,
            choice_text2: Report.DATE_DETAIL_MISSING_PITCHFX,
            choice_text3: Report.SINGLE_DATE_WITH_GAME_STATUS,
            f"{EMOJIS.get('BACK')} Return to Previous Menu": None,
        }
        return user_options_prompt(choices, PROMPT_TEXT)

    def get_date_range_report_type_from_user(self):
        choice_text1 = f"{MENU_NUMBERS.get(1)}  Dates Missing Data (Summary)"
        choice_text2 = f"{MENU_NUMBERS.get(2)}  All Dates In Range (Summary)"
        choice_text3 = f"{MENU_NUMBERS.get(3)}  Dates Missing Data (Detail)"
        choice_text4 = f"{MENU_NUMBERS.get(4)}  All Dates In Range (Detail)"
        choice_text5 = f"{MENU_NUMBERS.get(5)}  All Dates In Range + Missing PitchFx IDs (Detail)"
        choices = {
            choice_text1: Report.DATE_SUMMARY_MISSING_DATA,
            choice_text2: Report.DATE_SUMMARY_ALL_DATES,
            choice_text3: Report.DATE_DETAIL_MISSING_DATA,
            choice_text4: Report.DATE_DETAIL_ALL_DATES,
            choice_text5: Report.DATE_DETAIL_MISSING_PITCHFX,
            f"{EMOJIS.get('BACK')} Return to Previous Menu": None,
        }
        return user_options_prompt(choices, PROMPT_TEXT)

    def season_status_report(self):
        result = self.get_mlb_season_from_user()
        if result.failure:
            return result
        year = result.value
        result = self.get_season_report_type_from_user()
        if result.failure:
            return result
        report_type = result.value
        subprocess.run(["clear"])
        return report_season_status(self.db_session, year, report_type)

    def single_date_report(self):
        game_date = single_date_prompt("Report status for date:")
        result = self.get_single_date_report_type_from_user()
        if result.failure:
            return result
        report_type = result.value
        subprocess.run(["clear"])
        return report_status_single_date(self.db_session, game_date, report_type)

    def date_range_report(self):
        start_date = single_date_prompt("Report status start date: ")
        end_date = single_date_prompt("Report status end date: ")
        result = self.get_date_range_report_type_from_user()
        if result.failure:
            return result
        report_type = result.value
        subprocess.run(["clear"])
        return report_date_range_status(self.db_session, start_date, end_date, report_type)

    def get_mlb_season_from_user(self):
        result = season_prompt(self.db_session, "Select the MLB Season to report:")
        if result.failure:
            return result
        season = result.value
        return Result.Ok(season.year)



import subprocess
from abc import ABC, abstractmethod

from bullet import Bullet, colors, ScrollBar

from vigorish.cli.menu_item import MenuItem
from vigorish.util.result import Result
from vigorish.util.string_helpers import ellipsize


class Menu(MenuItem, ABC):
    menu_text = ""
    selected_menu_item_text = ""
    pointer = ""
    menu_items = []
    bullet_color = colors.bright(colors.foreground["default"])
    check_color = colors.bright(colors.foreground["default"])
    check_on_switch = colors.bright(colors.foreground["default"])
    pointer_color = colors.bright(colors.foreground["default"])

    def __init__(self, app):
        super().__init__(app)

    @property
    def menu_item_text_list(self):
        return list(self.menu_item_dict.keys())

    @property
    def menu_item_dict(self):
        return {ellipsize(item.menu_item_text, 70): item for item in self.menu_items}

    @property
    def menu_item_count(self):
        return len(self.menu_items)

    @property
    def selected_menu_item_pos(self):
        if not self.selected_menu_item_text:
            return 0
        menu_item_pos_dict = {ellipsize(item.menu_item_text, 70): i for i, item in enumerate(self.menu_items)}
        return menu_item_pos_dict.get(self.selected_menu_item_text, 0)

    @property
    def selected_menu_item(self):
        return self.menu_item_dict.get(self.selected_menu_item_text)

    def launch(self):
        exit_menu = False
        while not exit_menu:
            subprocess.run(["clear"])
            self.populate_menu_items()
            result = self.prompt_user_for_menu_selection()
            if result.failure:
                return result
            exit_menu = result.value
        return Result.Ok(exit_menu)

    def prompt_user_for_menu_selection(self):
        if self.menu_item_count <= 8:
            menu = self._get_bullet_menu()
            menu.pos = self.selected_menu_item_pos
        else:
            menu = self._get_scroll_menu()
        self.selected_menu_item_text = menu.launch()
        result = self.selected_menu_item.launch()
        exit_menu = self.selected_menu_item.exit_menu
        return Result.Ok(exit_menu) if result.success else result

    @abstractmethod
    def populate_menu_items(self):
        pass

    def _get_bullet_menu(self):
        return Bullet(
            self.menu_text,
            choices=self.menu_item_text_list,
            bullet=self.pointer,
            shift=1,
            indent=2,
            margin=2,
            bullet_color=self.bullet_color,
            background_color=self.background_color,
            background_on_switch=self.background_on_switch,
            word_color=self.word_color,
            word_on_switch=self.word_on_switch,
        )

    def _get_scroll_menu(self):
        return ScrollBar(
            self.menu_text,
            choices=self.menu_item_text_list,
            height=8,
            pointer=self.pointer,
            shift=1,
            indent=2,
            margin=2,
            pointer_color=self.pointer_color,
            background_color=self.background_color,
            background_on_switch=self.background_on_switch,
            word_color=self.word_color,
            word_on_switch=self.word_on_switch,
        )



"""Menu that contains other menus related to configuring/administrating vigorish."""
from vigorish.cli.menu import Menu
from vigorish.cli.menu_items.return_to_parent import ReturnToParent
from vigorish.cli.menus.config_settings_menu import ConfigSettingsMenu
from vigorish.cli.menus.env_var_settings_menu import EnvVarSettingsMenu
from vigorish.constants import EMOJIS


class SettingsMenu(Menu):
    def __init__(self, app):
        super().__init__(app)
        self.menu_text = "You can view/edit Config File settings or Env. Variable settings:"
        self.menu_item_text = "Settings"
        self.menu_item_emoji = EMOJIS.get("TOOLS", "")

    def populate_menu_items(self):
        self.menu_items = []
        self.menu_items.append(ConfigSettingsMenu(self.app))
        self.menu_items.append(EnvVarSettingsMenu(self.app))
        self.menu_items.append(ReturnToParent(self.app, "Return to Main Menu"))



"""decorators.log_call"""
import inspect
import logging
from datetime import datetime
from functools import wraps
from time import sleep

from vigorish.util.datetime_util import format_timedelta_str
from vigorish.util.dt_format_strings import DT_NAIVE


class RetryLimitExceededError(Exception):
    """Custom error raised by retry decorator when max_attempts have failed."""

    def __init__(self, func, max_attempts):
        error = f"Retry limit exceeded! (function: {func.__name__}, max attempts: {max_attempts})"
        super().__init__(error)


def retry(
    *,
    max_attempts=2,
    delay=1,
    exceptions=(Exception,),
    on_failure=None,
):
    """Retry the wrapped function when an exception is raised until max_attempts have failed."""

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            for remaining in reversed(range(max_attempts)):
                try:
                    return func(*args, **kwargs)
                except exceptions as ex:
                    if remaining <= 0:
                        raise RetryLimitExceededError(func, max_attempts) from ex
                    if on_failure:
                        on_failure(func, remaining, ex, delay)
                    sleep(delay)
                else:
                    break

        return wrapper

    return decorator


class LogCall:
    """Log call signature and execution time of decorated function."""

    def __init__(self, logger=None):
        self.logger = logger

    def __call__(self, func):
        if not self.logger:
            logging.basicConfig()
            self.logger = logging.getLogger(func.__module__)
            self.logger.setLevel(logging.INFO)

        @wraps(func)
        def wrapper(*args, **kwargs):
            func_call_args = get_function_call_args(func, *args, **kwargs)
            exec_start = datetime.now()
            result = func(*args, **kwargs)
            exec_finish = datetime.now()
            exec_time = format_timedelta_str(exec_finish - exec_start)
            exec_start_str = exec_start.strftime(DT_NAIVE)
            self.logger.info(f"{exec_start_str} | {func_call_args} | {exec_time}")
            return result

        def get_function_call_args(func, *args, **kwargs):
            """Return a string containing function name and list of all argument names/values."""
            func_args = inspect.signature(func).bind(*args, **kwargs)
            func_args.apply_defaults()
            func_args_str = ", ".join(f"{arg}={val}" for arg, val in func_args.arguments.items())
            return f"{func.__name__}({func_args_str})"

        return wrapper



import subprocess
from datetime import datetime, timezone
from pathlib import Path

from getch import pause
from halo import Halo

from vigorish.cli.components.prompts import user_options_prompt
from vigorish.cli.components.util import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_error,
    print_heading,
    shutdown_cli_immediately,
)
from vigorish.cli.components.viewers import DisplayPage, PageViewer
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.tasks import RestoreDatabaseTask
from vigorish.util.datetime_util import format_timedelta_str, get_local_utcoffset
from vigorish.util.dt_format_strings import DT_NAIVE, FILE_TIMESTAMP
from vigorish.util.result import Result


class RestoreDatabase(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.menu_item_text = "Restore Database from Backup"
        self.menu_item_emoji = EMOJIS.get("SPIRAL")
        self.exit_menu = False
        self.restore_db = RestoreDatabaseTask(self.app)
        self.backup_start_time = None
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.backup_folder = self.app.get_current_setting("DB_BACKUP_FOLDER_PATH")
        self.backup_zip_files = []
        self.table_count = 0
        self.total_tables = 0
        self.row_count = 0
        self.total_rows = 0

    def launch(self):
        result = self.find_db_backups()
        if result.failure:
            print_error(result.error)
            pause(message="\nPress any key to continue...")
            return Result.Ok(True)
        if not self.prompt_user_run_task():
            return Result.Ok(True)
        result = self.prompt_user_select_backup_to_restore()
        if result.failure:
            return Result.Ok(True)
        zip_file = result.value
        self.subscribe_to_events()
        result = self.restore_db.execute(zip_file=zip_file)
        self.unsubscribe_from_events()
        if result.failure:
            self.spinner.stop()
            return result
        pause(message="\nPress any key to continue...")
        shutdown_cli_immediately()

    def find_db_backups(self):
        self.backup_zip_files = list(Path(self.backup_folder).glob("*.zip"))
        if not self.backup_zip_files:
            return Result.Fail(f"Error! No backups found in folder:\n{self.backup_folder}")
        self.backup_zip_files.sort(key=lambda x: x.name, reverse=True)
        return Result.Ok()

    def prompt_user_run_task(self):
        subprocess.run(["clear"])
        heading = "Restore Database from Backup"
        prompt = "Would you like to run this task and restore the database?"
        pages = [DisplayPage(page, heading) for page in self.get_task_description_pages()]
        page_viewer = PageViewer(pages, prompt=prompt, text_color="bright_yellow", heading_color="bright_yellow")
        return page_viewer.launch()

    def prompt_user_select_backup_to_restore(self):
        subprocess.run(["clear"])
        print_heading("Restore Database from Backup", fg="bright_yellow")
        choices = {
            f"{MENU_NUMBERS.get(num,num)}  {self.get_backup_time(zip_file.stem)}": zip_file
            for num, zip_file in enumerate(self.backup_zip_files, start=1)
        }
        choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
        prompt = "Select a backup to restore (time shown is when the backup process was started):"
        return user_options_prompt(choices, prompt, clear_screen=False)

    def get_backup_time(self, timestamp):
        backup_created = datetime.strptime(timestamp, FILE_TIMESTAMP).replace(tzinfo=timezone.utc)
        time_span = format_timedelta_str(datetime.now(timezone.utc) - backup_created, precise=False)
        backup_created_aware = backup_created.astimezone(get_local_utcoffset())
        return f"{backup_created_aware.strftime(DT_NAIVE)} ({time_span} ago)"

    def prepare_database(self):
        subprocess.run(["clear"])
        print_heading("Restore Database from Backup", fg="bright_yellow")
        return self.app.prepare_database_for_restore()

    def unzip_backup_files_start(self):
        subprocess.run(["clear"])
        print_heading("Restore Database from Backup", fg="bright_yellow")
        self.backup_start_time = datetime.now()
        self.spinner.start()
        self.spinner.text = "Unzipping backup CSV files..."

    def unzip_backup_files_complete(self):
        self.spinner.stop()
        self.spinner = None

    def restore_database_start(self):
        subprocess.run(["clear"])
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.spinner.start()

    def restore_table_start(self, table):
        self.current_table = table
        self.spinner.text = self.get_spinner_text(0)

    def batch_insert_performed(self, batch_count):
        self.spinner.text = self.get_spinner_text(batch_count)

    def get_spinner_text(self, batch_count):
        return f"Restoring {self.current_table.__name__} table ({batch_count} Batches Added)..."

    def restore_database_complete(self):
        elapsed = format_timedelta_str(datetime.now() - self.backup_start_time)
        self.spinner.succeed(f"Successfully restored database from backup! (elapsed: {elapsed})")

    def subscribe_to_events(self):
        self.restore_db.events.unzip_backup_files_start += self.unzip_backup_files_start
        self.restore_db.events.unzip_backup_files_complete += self.unzip_backup_files_complete
        self.restore_db.events.restore_database_start += self.restore_database_start
        self.restore_db.events.restore_table_start += self.restore_table_start
        self.restore_db.events.batch_insert_performed += self.batch_insert_performed
        self.restore_db.events.restore_database_complete += self.restore_database_complete

    def unsubscribe_from_events(self):
        self.restore_db.events.unzip_backup_files_start -= self.unzip_backup_files_start
        self.restore_db.events.restore_table_start -= self.restore_table_start
        self.restore_db.events.batch_insert_performed -= self.batch_insert_performed
        self.restore_db.events.restore_database_complete -= self.restore_database_complete

    def get_task_description_pages(self):
        return [
            [
                (
                    "This task restores database tables that track the progress made scraping all "
                    "data sets for all MLB seasons.\n"
                ),
                ("Backup files are stored in the folder specified by the DB_BACKUP_FOLDER_PATH config setting."),
            ]
        ]



"""Calculate average time between pitches during at bat, between at bats and between innings."""
from datetime import datetime, timezone

from events import Events
from scipy.stats.mstats import winsorize

import vigorish.database as db
from vigorish.tasks.base import Task
from vigorish.util.datetime_util import TIME_ZONE_NEW_YORK
from vigorish.util.regex import PFX_TIMESTAMP_REGEX
from vigorish.util.result import Result
from vigorish.util.string_helpers import validate_bbref_game_id


class CalculateAvgPitchTimesTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self.events = Events(
            (
                "find_eligible_games_start",
                "find_eligible_games_complete",
                "calculate_pitch_metrics_start",
                "calculate_pitch_metrics_progress",
                "calculate_pitch_metrics_complete",
            )
        )

    def execute(self, trim_data_sets=True):
        self.events.find_eligible_games_start()
        game_ids = db.Season_Game_PitchApp_View.get_all_bbref_game_ids_combined_no_missing_pfx(self.db_engine)
        if not game_ids:
            return Result.Fail("No games meet the requirements for this process.")
        self.events.find_eligible_games_complete(game_ids)
        self.events.calculate_pitch_metrics_start()
        pitch_samples = []
        at_bat_samples = []
        inning_samples = []
        for num, game_id in enumerate(game_ids, start=1):
            combined_data = self.scraped_data.get_combined_game_data(game_id)
            if not combined_data:
                continue
            result = self.calc_pitch_metrics(combined_data)
            pitch_samples.extend(result[0])
            at_bat_samples.extend(result[2])
            inning_samples.extend(result[4])
            self.events.calculate_pitch_metrics_progress(num)
        self.events.calculate_pitch_metrics_complete()
        metrics = {
            "time_between_pitches": self.process_data_set(pitch_samples, trim=trim_data_sets),
            "time_between_at_bats": self.process_data_set(at_bat_samples, trim=trim_data_sets),
            "time_between_innings": self.process_data_set(inning_samples, trim=trim_data_sets),
        }
        return Result.Ok(metrics)

    def calc_pitch_metrics(self, combined_data):
        pitch_samples = []
        at_bat_samples = []
        inning_samples = []
        prev_ab_pfx = None
        prev_inn_pfx = None
        for half_inning in combined_data["play_by_play_data"]:
            for ab_num, at_bat in enumerate(half_inning["inning_events"], start=1):
                if (
                    at_bat["at_bat_pitchfx_audit"]["pitchfx_error"]
                    or at_bat["at_bat_pitchfx_audit"]["missing_pitchfx_count"]
                ):
                    continue
                for count, pfx in enumerate(at_bat["pitchfx"], start=1):
                    if count == 1 and prev_inn_pfx and ab_num == 1:
                        between_innings = self.get_seconds_between_pitches(prev_inn_pfx, pfx)
                        if between_innings > 0:
                            inning_samples.append(between_innings)
                        prev_inn_pfx = None
                    if count == 1 and prev_ab_pfx and ab_num != 1:
                        between_at_bats = self.get_seconds_between_pitches(prev_ab_pfx, pfx)
                        if between_at_bats > 0:
                            at_bat_samples.append(between_at_bats)
                        prev_ab_pfx = None
                    if count == len(at_bat["pitchfx"]):
                        if ab_num == len(half_inning["inning_events"]):
                            prev_inn_pfx = pfx
                        else:
                            prev_ab_pfx = pfx
                        continue
                    between_pitches = self.get_seconds_between_pitches(pfx, at_bat["pitchfx"][count])
                    if between_pitches > 0:
                        pitch_samples.append(between_pitches)
        return (
            pitch_samples,
            self.process_data_set(pitch_samples),
            at_bat_samples,
            self.process_data_set(at_bat_samples),
            inning_samples,
            self.process_data_set(inning_samples),
        )

    def get_seconds_between_pitches(self, pitch1, pitch2):
        pitch1_thrown = self.get_time_pitch_thrown(pitch1)
        pitch2_thrown = self.get_time_pitch_thrown(pitch2)
        if not pitch1_thrown or not pitch2_thrown:
            return 0
        return (pitch2_thrown - pitch1_thrown).total_seconds()

    def get_time_pitch_thrown(self, pfx):
        match = PFX_TIMESTAMP_REGEX.match(pfx["park_sv_id"])
        if not match:
            return None
        group_dict = match.groupdict()
        game_dict = validate_bbref_game_id(pfx["bbref_game_id"]).value
        try:
            timestamp = datetime(
                game_dict["game_date"].year,
                int(group_dict["month"]),
                int(group_dict["day"]),
                int(group_dict["hour"]),
                int(group_dict["minute"]),
                int(group_dict["second"]),
            )
        except ValueError:
            return None
        return timestamp.replace(tzinfo=timezone.utc).astimezone(TIME_ZONE_NEW_YORK)

    def process_data_set(self, data_set, trim=False):
        if not data_set or not isinstance(data_set, list):
            return {"error": "Data set is empty or is not a valid list"}
        if trim:
            data_set = winsorize(data_set, limits=[0.01, 0.01], nan_policy="omit")
        return {
            "total": int(sum(data_set)),
            "count": int(len(data_set)),
            "avg": round(sum(data_set) / len(data_set), 1),
            "max": int(max(data_set)),
            "min": int(min(data_set)),
            "range": int(max(data_set) - min(data_set)),
            "trim": trim,
        }



"""Errors encountered while executing scrape jobs."""
from datetime import timezone

from sqlalchemy import Boolean, Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.types import Enum

import vigorish.database as db
from vigorish.enums import DataSet
from vigorish.util.datetime_util import (
    get_local_utcoffset,
    localized_dt_string,
    make_tzaware,
    utc_now,
)


class ScrapeError(db.Base):
    """Errors encountered while executing scrape jobs."""

    __tablename__ = "scrape_error"
    id = Column(Integer, primary_key=True)
    occurred_at = Column(DateTime, default=utc_now)
    data_set = Column(Enum(DataSet), nullable=False)
    error_message = Column(String, nullable=False)
    fixed = Column(Boolean, default=False)
    job_id = Column(Integer, ForeignKey("scrape_job.id"))

    @hybrid_property
    def occurred_at_str(self):
        occurred_at_utc = make_tzaware(self.occurred_at, use_tz=timezone.utc, localize=False)
        return localized_dt_string(occurred_at_utc, use_tz=get_local_utcoffset())

    def __repr__(self):
        return f"<ScrapeError job_name={self.job.name}, job_id={self.job_id}, message={self.error_message}>"

    def __str__(self):
        return f"{self.occurred_at_str} | {self.error_message}"

    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}



from sqlalchemy import case, cast, Float, func

import vigorish.database as db

zone_rate_numer = func.sum(db.PitchFx.inside_strike_zone)
zone_rate_denom = func.count(db.PitchFx.id)
zone_rate = zone_rate_numer / cast(zone_rate_denom, Float)
zone_rate = case([(zone_rate_denom > 0, zone_rate)], else_=0.0).label("zone_rate")

cs_rate_numer = func.sum(db.PitchFx.called_strike)
cs_rate_denom = func.count(db.PitchFx.id)
called_strike_rate = cs_rate_numer / cast(cs_rate_denom, Float)
called_strike_rate = case([(cs_rate_denom > 0, called_strike_rate)], else_=0.0).label("called_strike_rate")

ss_rate_numer = func.sum(db.PitchFx.swinging_strike)
ss_rate_denom = func.count(db.PitchFx.id)
swinging_strike_rate = ss_rate_numer / cast(ss_rate_denom, Float)
swinging_strike_rate = case([(ss_rate_denom > 0, swinging_strike_rate)], else_=0.0).label("swinging_strike_rate")

whiff_rate_numer = func.sum(db.PitchFx.swinging_strike)
whiff_rate_denom = func.sum(db.PitchFx.batter_did_swing)
whiff_rate = whiff_rate_numer / cast(whiff_rate_denom, Float)
whiff_rate = case([(whiff_rate_denom > 0, whiff_rate)], else_=0.0).label("whiff_rate")

csw_rate_numer = func.sum(db.PitchFx.swinging_strike) + func.sum(db.PitchFx.called_strike)
csw_rate_denom = func.count(db.PitchFx.id)
csw_rate = csw_rate_numer / cast(csw_rate_denom, Float)
csw_rate = case([(csw_rate_denom > 0, csw_rate)], else_=0.0).label("csw_rate")

os_rate_numer = func.sum(db.PitchFx.swing_outside_zone)
os_rate_denom = func.sum(db.PitchFx.outside_strike_zone)
o_swing_rate_ = os_rate_numer / cast(os_rate_denom, Float)
o_swing_rate = case([(os_rate_denom > 0, o_swing_rate_)], else_=0.0).label("o_swing_rate")

zs_rate_numer = func.sum(db.PitchFx.swing_inside_zone)
zs_rate_denom = func.sum(db.PitchFx.inside_strike_zone)
z_swing_rate = zs_rate_numer / cast(zs_rate_denom, Float)
z_swing_rate = case([(zs_rate_denom > 0, z_swing_rate)], else_=0.0).label("z_swing_rate")

sw_rate_numer = func.sum(db.PitchFx.batter_did_swing)
sw_rate_denom = func.count(db.PitchFx.id)
swing_rate = sw_rate_numer / cast(sw_rate_denom, Float)
swing_rate = case([(sw_rate_denom > 0, swing_rate)], else_=0.0).label("swing_rate")

oc_rate_numer = func.sum(db.PitchFx.contact_outside_zone)
oc_rate_denom = func.sum(db.PitchFx.swing_outside_zone)
o_contact_rate = oc_rate_numer / cast(oc_rate_denom, Float)
o_contact_rate = case([(oc_rate_denom > 0, o_contact_rate)], else_=0.0).label("o_contact_rate")

zc_rate_numer = func.sum(db.PitchFx.contact_inside_zone)
zc_rate_denom = func.sum(db.PitchFx.swing_inside_zone)
z_contact_rate = zc_rate_numer / cast(zc_rate_denom, Float)
z_contact_rate = case([(zc_rate_denom > 0, z_contact_rate)], else_=0.0).label("z_contact_rate")

contact_rate_numer = func.sum(db.PitchFx.batter_made_contact)
contact_rate_denom = func.sum(db.PitchFx.batter_did_swing)
contact_rate = contact_rate_numer / cast(contact_rate_denom, Float)
contact_rate = case([(contact_rate_denom > 0, contact_rate)], else_=0.0).label("contact_rate")

gb_rate_numer = func.sum(db.PitchFx.is_ground_ball)
gb_rate_denom = func.sum(db.PitchFx.is_batted_ball)
ground_ball_rate = gb_rate_numer / cast(gb_rate_denom, Float)
ground_ball_rate = case([(gb_rate_denom > 0, ground_ball_rate)], else_=0.0).label("ground_ball_rate")

fb_rate_numer = func.sum(db.PitchFx.is_fly_ball)
fb_rate_denom = func.sum(db.PitchFx.is_batted_ball)
fly_ball_rate = fb_rate_numer / cast(fb_rate_denom, Float)
fly_ball_rate = case([(fb_rate_denom > 0, fly_ball_rate)], else_=0.0).label("fly_ball_rate")

ld_rate_numer = func.sum(db.PitchFx.is_line_drive)
ld_rate_denom = func.sum(db.PitchFx.is_batted_ball)
line_drive_rate = ld_rate_numer / cast(ld_rate_denom, Float)
line_drive_rate = case([(ld_rate_denom > 0, line_drive_rate)], else_=0.0).label("line_drive_rate")

pu_rate_numer = func.sum(db.PitchFx.is_pop_up)
pu_rate_denom = func.sum(db.PitchFx.is_batted_ball)
pop_up_rate = pu_rate_numer / cast(pu_rate_denom, Float)
pop_up_rate = case([(pu_rate_denom > 0, pop_up_rate)], else_=0.0).label("pop_up_rate")

total_hits = func.sum(db.PitchFx.ab_result_hit)
total_at_bats = (
    func.sum(db.PitchFx.ab_result_hit)
    + func.sum(db.PitchFx.ab_result_out)
    + func.sum(db.PitchFx.ab_result_error)
    - func.sum(db.PitchFx.ab_result_sac_hit)
    - func.sum(db.PitchFx.ab_result_sac_fly)
).label("total_at_bats")
avg_ = total_hits / cast(total_at_bats, Float)
avg = case([(total_at_bats > 0, avg_)], else_=0.0).label("avg")

obp_numerator = total_hits + func.sum(db.PitchFx.ab_result_bb) + func.sum(db.PitchFx.ab_result_hbp)
obp_denominator = (
    total_at_bats
    + func.sum(db.PitchFx.ab_result_bb)
    + func.sum(db.PitchFx.ab_result_hbp)
    + func.sum(db.PitchFx.ab_result_sac_fly)
)
obp_ = obp_numerator / cast(obp_denominator, Float)
obp = case([(obp_denominator > 0, obp_)], else_=0.0).label("obp")

slg_numerator = (
    func.sum(db.PitchFx.ab_result_single)
    + (func.sum(db.PitchFx.ab_result_double) * 2)
    + (func.sum(db.PitchFx.ab_result_triple) * 3)
    + (func.sum(db.PitchFx.ab_result_homerun) * 4)
)
slg_denominator = total_at_bats
slg_ = slg_numerator / cast(slg_denominator, Float)
slg = case([(slg_denominator > 0, slg_)], else_=0.0).label("slg")

ops = (obp + slg).label("ops")
iso = (slg - avg).label("iso")

bb_rate_numerator = func.sum(db.PitchFx.ab_result_bb)
bb_rate_denominator = func.sum(db.PitchFx.is_final_pitch_of_ab)
bb_rate_ = bb_rate_numerator / cast(bb_rate_denominator, Float)
bb_rate = case([(bb_rate_denominator > 0, bb_rate_)], else_=0.0).label("bb_rate")

k_rate_numerator = func.sum(db.PitchFx.ab_result_k)
k_rate_denominator = func.sum(db.PitchFx.is_final_pitch_of_ab)
k_rate_ = k_rate_numerator / cast(k_rate_denominator, Float)
k_rate = case([(k_rate_denominator > 0, k_rate_)], else_=0.0).label("k_rate")

hr_per_fb_numerator = func.sum(db.PitchFx.ab_result_homerun)
hr_per_fb_denominator = func.sum(db.PitchFx.is_fly_ball)
hr_per_fb_ = hr_per_fb_numerator / cast(hr_per_fb_denominator, Float)
hr_per_fb = case([(hr_per_fb_denominator > 0, hr_per_fb_)], else_=0.0).label("hr_per_fb")



from sqlalchemy import Boolean, Column, Float, Integer, String

import vigorish.database as db


class PitchTypePercentile(db.Base):

    __tablename__ = "pitch_type_percentiles"
    id = Column(Integer, primary_key=True)
    stat_name = Column(String)
    pitch_type = Column(String)
    thrown_r = Column(Boolean)
    thrown_l = Column(Boolean)
    thrown_both = Column(Boolean)
    percentile = Column(Float)
    stat_value = Column(Float)

    @classmethod
    def get_percentile_for_rhp_pos_stat(cls, db_session, stat_name, pfx_metrics):
        stat_value = getattr(pfx_metrics, stat_name)
        pt_percentile = (
            db_session.query(cls)
            .filter_by(stat_name=stat_name)
            .filter_by(thrown_r=True)
            .filter_by(pitch_type=str(pfx_metrics.pitch_type))
            .filter(cls.stat_value >= stat_value)
            .first()
        )
        return (stat_value, pt_percentile.percentile) if pt_percentile else (stat_value, 99.0)

    @classmethod
    def get_percentile_for_lhp_pos_stat(cls, db_session, stat_name, pfx_metrics):
        stat_value = getattr(pfx_metrics, stat_name)
        pt_percentile = (
            db_session.query(cls)
            .filter_by(stat_name=stat_name)
            .filter_by(thrown_l=True)
            .filter_by(pitch_type=str(pfx_metrics.pitch_type))
            .filter(cls.stat_value >= stat_value)
            .first()
        )
        return (stat_value, pt_percentile.percentile) if pt_percentile else (stat_value, 99.0)

    @classmethod
    def get_percentile_for_rhp_neg_stat(cls, db_session, stat_name, pfx_metrics):
        stat_value = getattr(pfx_metrics, stat_name)
        pt_percentile = (
            db_session.query(cls)
            .filter_by(stat_name=stat_name)
            .filter_by(thrown_r=True)
            .filter_by(pitch_type=str(pfx_metrics.pitch_type))
            .filter(cls.stat_value <= stat_value)
            .first()
        )
        return (stat_value, pt_percentile.percentile) if pt_percentile else (stat_value, 1.0)

    @classmethod
    def get_percentile_for_lhp_neg_stat(cls, db_session, stat_name, pfx_metrics):
        stat_value = getattr(pfx_metrics, stat_name)
        pt_percentile = (
            db_session.query(cls)
            .filter_by(stat_name=stat_name)
            .filter_by(thrown_l=True)
            .filter_by(pitch_type=str(pfx_metrics.pitch_type))
            .filter(cls.stat_value <= stat_value)
            .first()
        )
        return (stat_value, pt_percentile.percentile) if pt_percentile else (stat_value, 1.0)

    @classmethod
    def calculate_pitch_type_percentiles(cls, db_session, p_throws, pfx):
        return (
            cls.calculate_pitch_type_percentiles_for_rhp(db_session, pfx)
            if p_throws == "R"
            else cls.calculate_pitch_type_percentiles_for_lhp(db_session, pfx)
        )

    @classmethod
    def calculate_pitch_type_percentiles_for_rhp(cls, db_session, pfx):
        return {
            "pitch_type": pfx.pitch_type,
            "avg_speed": cls.get_percentile_for_rhp_pos_stat(db_session, "avg_speed", pfx),
            "ops": cls.get_percentile_for_rhp_neg_stat(db_session, "ops", pfx),
            "whiff_rate": cls.get_percentile_for_rhp_pos_stat(db_session, "whiff_rate", pfx),
            "zone_rate": cls.get_percentile_for_rhp_pos_stat(db_session, "zone_rate", pfx),
            "contact_rate": cls.get_percentile_for_rhp_neg_stat(db_session, "contact_rate", pfx),
            "o_swing_rate": cls.get_percentile_for_rhp_pos_stat(db_session, "o_swing_rate", pfx),
        }

    @classmethod
    def calculate_pitch_type_percentiles_for_lhp(cls, db_session, pfx):
        return {
            "pitch_type": pfx.pitch_type,
            "avg_speed": cls.get_percentile_for_lhp_pos_stat(db_session, "avg_speed", pfx),
            "ops": cls.get_percentile_for_lhp_neg_stat(db_session, "ops", pfx),
            "whiff_rate": cls.get_percentile_for_lhp_pos_stat(db_session, "whiff_rate", pfx),
            "zone_rate": cls.get_percentile_for_lhp_pos_stat(db_session, "zone_rate", pfx),
            "contact_rate": cls.get_percentile_for_lhp_neg_stat(db_session, "contact_rate", pfx),
            "o_swing_rate": cls.get_percentile_for_lhp_pos_stat(db_session, "o_swing_rate", pfx),
        }



from dataclasses import dataclass

from vigorish.cli.components.util import print_heading, print_message


@dataclass(frozen=True)
class DisplayTable:
    table: str
    heading: str = None
    message: str = None

    def display(self, heading_color=None, message_color=None, table_color=None):
        if self.heading:
            print_heading(self.heading, fg=heading_color)
        if self.message:
            print_message(self.message, fg=message_color, wrap=False)
            print()
        print_message(self.table, fg=table_color, wrap=False)



"""Menu item that combines scraped boxscore data and pitchfx data for a single game."""
import logging
import subprocess
import time
from collections import defaultdict

import enlighten
from getch import pause
from halo import Halo
from tabulate import tabulate

import vigorish.database as db
from vigorish.cli.components import (
    audit_report_season_prompt,
    get_random_cli_color,
    get_random_dots_spinner,
    print_message,
    user_options_prompt,
)
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.enums import AuditError, ScrapeCondition
from vigorish.tasks import CombineScrapedDataTask
from vigorish.util.dt_format_strings import DATE_MONTH_NAME
from vigorish.util.list_helpers import flatten_list2d
from vigorish.util.result import Result
from vigorish.util.string_helpers import validate_bbref_game_id, validate_pitch_app_id

STATUS_BAR_FORMAT = (
    "{total_combined}/{total_games} Games Combined"
    "{fill}"
    "Combining Data for Games Played On: {date_str}"
    "{fill}"
    "Elapsed: {elapsed}"
)
STATUS_BAR_COLOR = "bold_gray100_on_darkviolet"
DATE_BAR_FORMAT = "{desc}{desc_pad}{percentage:3.0f}% |{bar}| {count:{len_total}d}/{total:d} {unit}"

logging.basicConfig(level=logging.INFO)
LOGGER = logging.getLogger(__name__)


class CombineScrapedData(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.combine_data = CombineScrapedDataTask(app)
        self._season = None
        self._date_game_id_map = {}
        self.pbar_manager = enlighten.get_manager()
        self.menu_item_text = "Combine Game Data"
        self.menu_item_emoji = EMOJIS.get("BANG", "")
        self.exit_menu = False
        self.combine_condition = self.app.get_current_setting("SCRAPED_DATA_COMBINE_CONDITION")
        self.audit_type = None
        self.scrape_year = None
        self.current_game_date = None
        self.current_game_id = None
        self.combine_data_success_game_ids = []
        self.combine_data_fail_results = {}
        self.all_pfx_errors = defaultdict(dict)
        self.status_bar = None
        self.date_progress_bar = None
        self.game_progress_bar_success = None
        self.game_progress_bar_fail = None
        self.game_progress_bar_error = None

    @property
    def terminal(self):
        return self.pbar_manager.term if self.pbar_manager else None

    @property
    def game_bar_format(self):
        SUCCESS = EMOJIS.get("PASSED", "")
        FAIL = EMOJIS.get("FAILED", "")
        ERROR = EMOJIS.get("CONFUSED", "")
        return (
            "{desc}{desc_pad}{percentage:3.0f}% |{bar}| "
            + self.terminal.green3(f"{SUCCESS}" + " {count_0:{len_total}d}")
            + " | "
            + self.terminal.red2(f"{FAIL}" + " {count_1:{len_total}d}")
            + " | "
            + self.terminal.yellow2(f"{ERROR}" + " {count_2:{len_total}d}")
            + " | ETA: {eta}"
        )

    @property
    def season(self):
        if not self.scrape_year:
            return None
        if self._season:
            return self._season
        self._season = db.Season.find_by_year(self.db_session, self.scrape_year)
        return self._season

    @property
    def date_game_id_map(self):
        if not self.scrape_year:
            return {}
        if self._date_game_id_map:
            return self._date_game_id_map
        scraped_game_date_map = self.get_scraped_game_date_map(self.scrape_year)
        self._date_game_id_map = defaultdict(list)
        for game_map in scraped_game_date_map:
            self._date_game_id_map[game_map["game_date"]].append(game_map["game_id"])
        return self._date_game_id_map

    @property
    def all_dates_in_season(self):
        return [game_date.date() for game_date in self.season.get_date_range()] if self.season else None

    @property
    def all_dates_with_eligible_games(self):
        return sorted(set(self.date_game_id_map.keys()))

    @property
    def total_dates(self):
        return len(self.all_dates_in_season) if self.audit_type == "SEASON" else 1

    @property
    def all_eligible_games_in_season(self):
        return sorted(flatten_list2d(list(self.date_game_id_map.values())))

    @property
    def total_games(self):
        return (
            sum(len(game_ids) for game_ids in self.date_game_id_map.values())
            if self.audit_type == "SEASON"
            else len(self.date_game_id_map.get(self.current_game_date, None))
        )

    @property
    def total_combined_success(self):
        return len(self.combine_data_success_game_ids)

    @property
    def total_combined_fail(self):
        return len(self.combine_data_fail_results)

    @property
    def total_combined(self):
        return (
            self.total_combined_success
            + self.total_combined_fail
            + self.total_games_invalid_pitchfx
            + self.total_games_pitchfx_error
        )

    @property
    def pfx_errors(self):
        return self.all_pfx_errors[AuditError.PITCHFX_ERROR]

    @property
    def invalid_pfx(self):
        return self.all_pfx_errors[AuditError.INVALID_PITCHFX_DATA]

    @property
    def total_games_pitchfx_error(self):
        return len(self.pfx_errors)

    @property
    def total_games_invalid_pitchfx(self):
        return len(self.invalid_pfx)

    @property
    def total_games_any_pfx_error(self):
        return len(self.pfx_errors) + len(self.invalid_pfx)

    @property
    def game_ids_any_pitchfx_error(self):
        unique_game_ids = {game_id for error_dict in self.all_pfx_errors.values() for game_id in error_dict.keys()}
        return list(unique_game_ids)

    @property
    def pitch_apps_pitchfx_error(self):
        return flatten_list2d(list(pitch_app_dict.keys()) for pitch_app_dict in self.pfx_errors.values())

    @property
    def pitch_apps_invalid_pitchfx(self):
        return flatten_list2d(list(pitch_app_dict.keys()) for pitch_app_dict in self.invalid_pfx.values())

    @property
    def pitch_apps_any_pfx_error(self):
        return sorted(set(self.pitch_apps_pitchfx_error + self.pitch_apps_invalid_pitchfx))

    @property
    def total_pitch_apps_pitchfx_error(self):
        return len(self.pitch_apps_pitchfx_error)

    @property
    def total_pitch_apps_invalid_pitchfx(self):
        return len(self.pitch_apps_invalid_pitchfx)

    @property
    def total_pitch_apps_any_pitchfx_error(self):
        return len(self.pitch_apps_pitchfx_error) + len(self.pitch_apps_invalid_pitchfx)

    @property
    def game_id_pitch_app_id_map(self):
        id_map = defaultdict(list)
        for pitch_app_id in self.pitch_apps_any_pfx_error:
            result = validate_pitch_app_id(pitch_app_id)
            if result.failure:
                return {}
            pitch_app_dict = result.value
            id_map[pitch_app_dict["game_id"]].append(pitch_app_id)
        return id_map

    @property
    def at_bats_pitchfx_error(self):
        return flatten_list2d(pitch_app_dict.values() for pitch_app_dict in self.all_pfx_errors.values())

    @property
    def at_bats_invalid_pitchfx(self):
        return flatten_list2d(pitch_app_dict.values() for pitch_app_dict in self.invalid_pfx.values())

    @property
    def total_at_bats_pitchfx_error(self):
        return len(self.at_bats_pitchfx_error)

    @property
    def total_at_bats_invalid_pitchfx(self):
        return len(self.at_bats_invalid_pitchfx)

    @property
    def total_at_bats_any_pitchfx_error(self):
        return len(self.at_bats_pitchfx_error) + len(self.at_bats_invalid_pitchfx)

    @property
    def failed_game_ids(self):
        return list(self.combine_data_fail_results.keys())

    @property
    def combined_success_and_no_pfx_errors(self):
        return not self.total_combined_fail and not self.total_games_any_pfx_error

    def launch(self):
        subprocess.run(["clear"])
        exit_menu = False
        while not exit_menu:
            result = self.audit_type_prompt()
            if result.failure:
                exit_menu = True
                continue
            self.audit_type = result.value
            if self.audit_type == "SEASON":
                result = self.combine_games_for_season()
            if self.audit_type == "DATE":
                result = self.combine_games_for_date()
            if self.audit_type == "GAME":
                result = self.combine_single_game()
                if result.success:
                    exit_menu = True
                    continue
            if result.failure:
                continue
            self.display_results()
        return Result.Ok(self.exit_menu)

    def combine_games_for_season(self):
        result = audit_report_season_prompt(self.app.audit_report)
        if result.failure:
            return result
        self.scrape_year = result.value
        self.pbar_manager = enlighten.get_manager()
        self.init_progress_bars(game_date=self.all_dates_in_season[0])
        subprocess.run(["clear"])
        for game_date in self.all_dates_in_season:
            if self.every_eligible_game_is_combined():
                num_days_remaining = self.get_number_of_days_remaining()
                self.update_progress_bars(game_date)
                self.date_progress_bar.update(num_days_remaining)
                LOGGER.info(f"Processed all eligible games for MLB {self.scrape_year}.")
                time.sleep(1.5)
                break
            game_ids = self.date_game_id_map.get(game_date, None)
            if not game_ids:
                self.update_progress_bars(game_date)
                self.date_progress_bar.update()
                time.sleep(0.75)
                continue
            result = self.combine_selected_games(game_date, game_ids)
            if result.failure:
                return result
            self.date_progress_bar.update()
        self.close_progress_bars()
        return Result.Ok()

    def init_progress_bars(self, game_date):
        date_str = game_date.strftime(DATE_MONTH_NAME)
        self.status_bar = self.pbar_manager.status_bar(
            status_format=STATUS_BAR_FORMAT,
            color=STATUS_BAR_COLOR,
            justify=enlighten.Justify.CENTER,
            total_combined=self.total_combined,
            total_games=self.total_games,
            date_str=date_str,
        )
        self.date_progress_bar = self.pbar_manager.counter(
            total=self.total_dates,
            bar_format=DATE_BAR_FORMAT,
            desc=date_str,
            unit="days",
            color="cyan2",
            autorefresh=True,
        )
        self.game_progress_bar_success = self.pbar_manager.counter(
            total=self.total_games,
            bar_format=self.game_bar_format,
            desc="Combining data for game...",
            unit="games",
            color="green3",
            autorefresh=True,
        )
        self.game_progress_bar_fail = self.game_progress_bar_success.add_subcounter("red2")
        self.game_progress_bar_error = self.game_progress_bar_success.add_subcounter("yellow2")

    def every_eligible_game_is_combined(self):
        return self.total_combined == self.total_games

    def get_number_of_days_remaining(self):
        return self.total_dates - self.date_progress_bar.count

    def combine_selected_games(self, game_date, game_ids):
        for bbref_game_id in game_ids:
            fail_results = []
            self.current_game_id = bbref_game_id
            self.update_progress_bars(game_date)
            # LOGGER.info(f"Begin combining scraped data for game: {bbref_game_id}")
            result = self.combine_data.execute(bbref_game_id)
            if not result["gather_scraped_data_success"]:
                LOGGER.info(f"Unable to combine data for game: {bbref_game_id}")
                LOGGER.info(f"An error occurred gathering scraped data for game: {bbref_game_id}")
                return Result.Fail(result["error"])
            if not result["combined_data_success"]:
                self.combine_data_fail_results[bbref_game_id] = result["error"]
                self.game_progress_bar_fail.update()
                LOGGER.info(f"Failed to combine scraped data for game: {bbref_game_id}")
                LOGGER.info(result["error"])
                continue
            if not result["update_pitch_apps_success"]:
                LOGGER.info(f"Error occurred updating pitch apps for game: {bbref_game_id}")
                LOGGER.info(result["error"])
                return Result.Fail(result["error"])
            pfx_errors = result["results"]["pfx_errors"]
            if pfx_errors.get("pitchfx_error", []):
                self.pfx_errors[bbref_game_id] = pfx_errors["pitchfx_error"]
                fail_results.append(pfx_errors["pitchfx_error"])
            if pfx_errors.get("invalid_pitchfx", []):
                self.invalid_pfx[bbref_game_id] = pfx_errors["invalid_pitchfx"]
                fail_results.append(pfx_errors["invalid_pitchfx"])
            if fail_results:
                self.game_progress_bar_error.update()
                self.log_pfx_data_error_details(bbref_game_id, fail_results)
            else:
                self.combine_data_success_game_ids.append(bbref_game_id)
                self.game_progress_bar_success.update()
                # LOGGER.info(f"Successfully combined scraped data for game: {bbref_game_id}")
        return Result.Ok()

    def update_progress_bars(self, game_date):
        date_str = game_date.strftime(DATE_MONTH_NAME)
        self.status_bar.update(
            total_combined=self.total_combined,
            total_games=self.total_games,
            date_str=date_str,
        )
        self.date_progress_bar.desc = date_str
        self.game_progress_bar_success.desc = self.current_game_id

    def log_pfx_data_error_details(self, bbref_game_id, fail_results):
        total_pitch_apps = sum(len(f.keys()) for f in fail_results)
        pitch_apps_plural = "pitch appearances" if total_pitch_apps > 1 else "pitch appearance"
        total_at_bats = sum(len(at_bat_ids) for f in fail_results for at_bat_ids in f.values())
        at_bats_plural = "at bats" if total_at_bats > 1 else "at bat"
        LOGGER.info(f"PitchFX data could not be reconciled for game: {bbref_game_id}")
        LOGGER.info(
            f"{total_pitch_apps} {pitch_apps_plural} with data errors ({total_at_bats} total {at_bats_plural})\n"
        )

    def close_progress_bars(self):
        self.status_bar.close()
        self.date_progress_bar.close()
        self.game_progress_bar_success.close()
        self.pbar_manager.stop()

    def display_results(self):
        subprocess.run(["clear"])
        if self.combined_success_and_no_pfx_errors:
            plural = "games total" if self.total_games > 1 else "game"
            success_message = f"\nAll game data ({self.total_games} {plural}) combined, no errors"
            print_message(success_message, wrap=False, fg="bright_cyan", bold=True)
        if self.failed_game_ids:
            self.display_games_failed_to_combine()
        if self.all_pfx_errors:
            self.display_pitchfx_errors()
        pause(message="Press any key to continue...")

    def display_games_failed_to_combine(self):
        error_message = f"Error prevented scraped data being combined for {len(self.failed_game_ids)} games:"
        error_details = [
            {"bbref_game_id": game_id, "error": error} for game_id, error in self.combine_data_fail_results
        ]
        print_message(error_message, wrap=False, fg="bright_red", bold=True)
        print_message(tabulate(error_details, headers="keys"), wrap=False, fg="bright_red")

    def display_pitchfx_errors(self):
        games_plural = "games contain" if self.total_games_any_pfx_error > 1 else "game contains"
        ab_plural = "at bats" if self.total_at_bats_any_pitchfx_error > 1 else "at bat"
        message = (
            f"{self.total_games_any_pfx_error} {games_plural} invalid PitchFX data for a total "
            f"of {self.total_at_bats_any_pitchfx_error} {ab_plural}, you can view details "
            "of each at bat and attempt to fix these errors using the Investigate Failures menu."
        )
        print_message(message, fg="bright_cyan")
        print()

    def combine_games_for_date(self):
        result = audit_report_season_prompt(self.app.audit_report)
        if result.failure:
            return result
        self.scrape_year = result.value
        result = self.game_date_prompt()
        if result.failure:
            return result
        self.current_game_date = result.value
        self.pbar_manager = enlighten.get_manager()
        self.init_progress_bars(game_date=self.current_game_date)
        subprocess.run(["clear"])
        game_ids = self.date_game_id_map.get(self.current_game_date, None)
        if not game_ids:
            game_date_str = self.current_game_date.strftime(DATE_MONTH_NAME)
            message = f"All games on {game_date_str} have already been combined."
            print_message(message, fg="bright_cyan", bold=True)
            self.close_progress_bars()
            return Result.Ok()
        result = self.combine_selected_games(self.current_game_date, game_ids)
        self.date_progress_bar.update()
        self.close_progress_bars()
        return result

    def combine_single_game(self):
        result = audit_report_season_prompt(self.app.audit_report)
        if result.failure:
            return result
        self.scrape_year = result.value
        result = self.game_id_prompt()
        if result.failure:
            return result
        combine_game_id = result.value
        return self.combine_scraped_data_for_game(combine_game_id)

    def combine_scraped_data_for_game(self, combine_game_id):
        subprocess.run(["clear"])
        spinner = Halo(color=get_random_cli_color(), spinner=get_random_dots_spinner())
        spinner.text = f"Combining scraped data for {combine_game_id}..."
        spinner.start()
        result = self.combine_data.execute(combine_game_id)
        if not (
            result["gather_scraped_data_success"]
            and result["combined_data_success"]
            and result["update_pitch_apps_success"]
        ):
            spinner.fail(f"Failed to combine data for {combine_game_id}!")
            print_message(result["error"], wrap=False, fg="bright_red", bold=True)
            return Result.Fail(result["error"])
        spinner.stop()
        pfx_errors = result["results"]["pfx_errors"]
        if pfx_errors.get("pitchfx_error", []):
            self.pfx_errors[combine_game_id] = pfx_errors["pitchfx_error"]
        if pfx_errors.get("invalid_pitchfx", []):
            self.invalid_pfx[combine_game_id] = pfx_errors["invalid_pitchfx"]
        if self.total_pitch_apps_any_pitchfx_error > 0:
            pitch_apps_plural = (
                "pitch appearances" if self.total_pitch_apps_any_pitchfx_error > 1 else "pitch appearance"
            )
            at_bats_plural = "at bats" if self.total_at_bats_any_pitchfx_error > 1 else "at bat"
            message = (
                f"PitchFX data could not be reconciled for game: {combine_game_id}\n"
                f"{self.total_pitch_apps_any_pitchfx_error} {pitch_apps_plural} with data errors "
                f"({self.total_at_bats_any_pitchfx_error} total {at_bats_plural})\n"
            )
            print_message(message, fg="bright_yellow", bold=True)
        else:
            message = f"All scraped data for {combine_game_id} was successfully combined!"
            print_message(message, fg="bright_cyan", bold=True)
        pause(message="Press any key to continue...")
        return Result.Ok()

    def audit_type_prompt(self):
        prompt = (
            "For games where all data has been scraped, would you like to combine and audit the "
            "data for all games in the same season, all games on a single date or for a single "
            "game?"
        )
        choices = {
            f"{MENU_NUMBERS.get(1)}  By Season": "SEASON",
            f"{MENU_NUMBERS.get(2)}  By Date": "DATE",
            f"{MENU_NUMBERS.get(3)}  By Game": "GAME",
            f"{EMOJIS.get('BACK')} Return to Main Menu": None,
        }
        return user_options_prompt(choices, prompt)

    def game_date_prompt(self):
        choices = {
            f"{EMOJIS.get('BOLT')} {game_date.strftime(DATE_MONTH_NAME)}": game_date
            for game_date in self.all_dates_with_eligible_games
        }
        choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
        prompt = "Select a date to combine scraped data for all games that took place on that date"
        return user_options_prompt(choices, prompt)

    def game_id_prompt(self):
        choices = {f"{EMOJIS.get('BOLT')} {game_id}": game_id for game_id in self.all_eligible_games_in_season}
        choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
        prompt = "Select the BBRef.com Game ID to combine scraped data:"
        return user_options_prompt(choices, prompt)

    def get_scraped_game_date_map(self, year):
        return (
            self.get_never_audited_game_date_map(year)
            if self.combine_condition == ScrapeCondition.ONLY_MISSING_DATA
            else self.get_all_eligible_game_date_map(year)
            if self.combine_condition == ScrapeCondition.ALWAYS
            else None
        )

    def get_never_audited_game_date_map(self, year):
        return [self.get_game_id_date_map(gid) for gid in self.app.audit_report[year]["scraped"]]

    def get_successful_game_date_map(self, year):
        return [self.get_game_id_date_map(gid) for gid in self.app.audit_report[year]["successful"]]

    def get_failed_game_date_map(self, year):
        return [self.get_game_id_date_map(gid) for gid in self.app.audit_report[year]["failed"]]

    def get_all_eligible_game_date_map(self, year):
        return [self.get_game_id_date_map(gid) for gid in self.get_all_eligible_game_ids(year)]

    def get_all_eligible_game_ids(self, year):
        return (
            self.app.audit_report[year]["scraped"]
            + self.app.audit_report[year]["successful"]
            + self.app.audit_report[year]["failed"]
            + self.app.audit_report[year]["pfx_error"]
            + self.app.audit_report[year]["invalid_pfx"]
        )

    def get_game_id_date_map(self, bbref_game_id):
        game_dict = validate_bbref_game_id(bbref_game_id).value
        return {"game_id": bbref_game_id, "game_date": game_dict["game_date"], "count": 1}



from dacite import from_dict

from vigorish.enums import DataSet, VigFile
from vigorish.scrape.url_details import UrlDetails
from vigorish.util.datetime_util import get_date_range
from vigorish.util.dt_format_strings import DATE_MONTH_NAME
from vigorish.util.result import Result


def create_url_set(start_date, end_date, db_job, data_set, scraped_data):
    url_set = {}
    for game_date in get_date_range(start_date, end_date):
        result = create_url_set_for_date(db_job, data_set, scraped_data, game_date)
        if result.failure:
            return result
        url_set[game_date] = result.value
    return Result.Ok(url_set)


def create_url_set_for_date(db_job, data_set, scraped_data, game_date):
    create_url_set_dict = {
        DataSet.BROOKS_GAMES_FOR_DATE: create_url_for_brooks_games_for_date,
        DataSet.BROOKS_PITCH_LOGS: create_urls_for_brooks_pitch_logs_for_date,
        DataSet.BROOKS_PITCHFX: create_urls_for_brooks_pitchfx_logs_for_date,
        DataSet.BBREF_GAMES_FOR_DATE: create_url_for_bbref_games_for_date,
        DataSet.BBREF_BOXSCORES: create_urls_for_bbref_boxscores_for_date,
    }
    return create_url_set_dict[data_set](db_job, scraped_data, game_date)


def create_url_for_brooks_games_for_date(db_job, scraped_data, game_date):
    data_set = DataSet.BROOKS_GAMES_FOR_DATE
    url_data = {
        "url": get_url_for_brooks_games_for_date(game_date),
        "url_id": game_date,
        "fileName": get_filename(scraped_data, data_set, game_date),
        "cachedHtmlFolderPath": get_cached_html_folderpath(scraped_data, data_set, game_date),
        "scrapedHtmlFolderpath": get_scraped_html_folderpath(db_job, data_set),
    }
    return Result.Ok([from_dict(data_class=UrlDetails, data=url_data)])


def create_urls_for_brooks_pitch_logs_for_date(db_job, scraped_data, game_date):
    data_set = DataSet.BROOKS_PITCH_LOGS
    req_data_set = DataSet.BROOKS_GAMES_FOR_DATE
    games_for_date = scraped_data.get_brooks_games_for_date(game_date)
    if not games_for_date:
        return Result.Fail(get_unscraped_data_error(data_set, req_data_set, game_date))
    urls = []
    for game in games_for_date.games:
        if game.might_be_postponed:
            continue
        for pitcher_id, pitch_log_url in game.pitcher_appearance_dict.items():
            pitch_app_id = f"{game.bbref_game_id}_{pitcher_id}"
            url_data = {
                "url": pitch_log_url,
                "url_id": pitch_app_id,
                "fileName": get_filename(scraped_data, data_set, pitch_app_id),
                "cachedHtmlFolderPath": get_cached_html_folderpath(scraped_data, data_set, game_date),
                "scrapedHtmlFolderpath": get_scraped_html_folderpath(db_job, data_set),
            }
            urls.append(from_dict(data_class=UrlDetails, data=url_data))
    return Result.Ok(urls)


def create_urls_for_brooks_pitchfx_logs_for_date(db_job, scraped_data, game_date):
    data_set = DataSet.BROOKS_PITCHFX
    req_data_set_1 = DataSet.BROOKS_GAMES_FOR_DATE
    req_data_set_2 = DataSet.BROOKS_PITCH_LOGS
    games_for_date = scraped_data.get_brooks_games_for_date(game_date)
    if not games_for_date:
        return Result.Fail(get_unscraped_data_error(data_set, req_data_set_1, game_date))
    urls = []
    if not games_for_date.game_count:
        return Result.Ok(urls)
    pitch_logs_for_date = scraped_data.get_all_brooks_pitch_logs_for_date(game_date)
    if not pitch_logs_for_date:
        return Result.Fail(get_unscraped_data_error(data_set, req_data_set_2, game_date))
    for pitch_logs_for_game in pitch_logs_for_date:
        for pitch_log in pitch_logs_for_game.pitch_logs:
            if not pitch_log.parsed_all_info:
                continue
            pitch_app_id = f"{pitch_log.bbref_game_id}_{pitch_log.pitcher_id_mlb}"
            url_data = {
                "url": pitch_log.pitchfx_url,
                "url_id": pitch_app_id,
                "fileName": get_filename(scraped_data, data_set, pitch_app_id),
                "cachedHtmlFolderPath": get_cached_html_folderpath(scraped_data, data_set, game_date),
                "scrapedHtmlFolderpath": get_scraped_html_folderpath(db_job, data_set),
            }
            urls.append(from_dict(data_class=UrlDetails, data=url_data))
    return Result.Ok(urls)


def create_url_for_bbref_games_for_date(db_job, scraped_data, game_date):
    data_set = DataSet.BBREF_GAMES_FOR_DATE
    url_data = {
        "url": get_url_for_bbref_games_for_date(game_date),
        "url_id": game_date,
        "fileName": get_filename(scraped_data, data_set, game_date),
        "cachedHtmlFolderPath": get_cached_html_folderpath(scraped_data, data_set, game_date),
        "scrapedHtmlFolderpath": get_scraped_html_folderpath(db_job, data_set),
    }
    return Result.Ok([from_dict(data_class=UrlDetails, data=url_data)])


def create_urls_for_bbref_boxscores_for_date(db_job, scraped_data, game_date):
    data_set = DataSet.BBREF_BOXSCORES
    req_data_set = DataSet.BBREF_GAMES_FOR_DATE
    bbref_games_for_date = scraped_data.get_bbref_games_for_date(game_date)
    if not bbref_games_for_date:
        return Result.Fail(get_unscraped_data_error(data_set, req_data_set, game_date))
    urls = []
    for game_info in bbref_games_for_date.games:
        url_data = {
            "url": game_info.url,
            "url_id": game_info.bbref_game_id,
            "fileName": get_filename(scraped_data, data_set, game_info.bbref_game_id),
            "cachedHtmlFolderPath": get_cached_html_folderpath(scraped_data, data_set, game_date),
            "scrapedHtmlFolderpath": get_scraped_html_folderpath(db_job, data_set),
        }
        urls.append(from_dict(data_class=UrlDetails, data=url_data))
    return Result.Ok(urls)


def get_unscraped_data_error(data_set, req_data_set, game_date):
    return (
        f"Unable to create {data_set} URLs for {game_date.strftime(DATE_MONTH_NAME)} since "
        f"{req_data_set} for this date has not been scraped."
    )


def get_url_for_brooks_games_for_date(game_date):
    month = game_date.month
    day = game_date.day
    year = game_date.year
    return f"https://www.brooksbaseball.net/dashboard.php?dts={month}/{day}/{year}"


def get_url_for_bbref_games_for_date(game_date):
    month = game_date.month
    day = game_date.day
    year = game_date.year
    return f"https://www.baseball-reference.com/boxes/?month={month}&day={day}&year={year}"


def get_filename(scraped_data, data_set, identifier):
    return scraped_data.file_helper.filename_dict[VigFile.SCRAPED_HTML][data_set](identifier)


def get_cached_html_folderpath(scraped_data, data_set, game_date):
    return scraped_data.file_helper.get_local_folderpath(VigFile.SCRAPED_HTML, data_set, game_date)


def get_scraped_html_folderpath(db_job, data_set):
    folderpath = db_job.scraped_html_folders[data_set]
    return str(folderpath.resolve())
